#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import builtins
import datetime
import functools as ftz
import itertools as itz
import logging as logmod
import os
import re
import time
import types
import warnings
import weakref
from sys import stderr, stdout
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv import Mixin, Proto
from jgdv.structs.metalord.singleton import MLSingleton

# ##-- end 1st party imports

from . import _interface as API# noqa: N812
from .format import StripColourFormatter
from .logger import JGDVLogger
from .logger_spec import LoggerSpec

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, overload

if TYPE_CHECKING:
    from jgdv.structs.chainguard import ChainGuard
    import pathlib as pl
    import enum
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable
    from ._interface import Logger

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

ENV                  : dict              = os.environ # type: ignore
PRINTER_INITIAL_SPEC : Final[LoggerSpec] = LoggerSpec.build(API.default_printer)
INITIAL_SPEC         : Final[LoggerSpec] = LoggerSpec.build(API.default_stdout)

##--|

class PrintCapture_m:
    """ Mixin for redirecting builtins.print to a file  """
    original_print : Maybe[Callable]

    def capture_printing_to_file(self, path:str|pl.Path=API.default_print_file, *, disable_warning:bool=False) -> None:
        """ Modifies builtins.print to also print to a file

        Setup a file handler for a separate logger,
        to keep a trace of anything printed.
        Strips colour print command codes out of any string
        printed strings are logged at DEBUG level
        """
        match getattr(self, "original_print", None):
            case None:
                self.original_print = builtins.print
            case _:
                return

        if not disable_warning:
            warnings.warn("Modifying builtins.print", RuntimeWarning, 2)

        file_handler = logmod.FileHandler(path, mode='w')
        file_handler.setLevel(logmod.DEBUG)
        file_handler.setFormatter(StripColourFormatter())

        print_logger = logmod.getLogger(f"{API.PRINTER_NAME}.intercept")
        print_logger.setLevel(logmod.DEBUG)
        print_logger.addHandler(file_handler)
        print_logger.propagate = False

        @ftz.wraps(self.original_print)
        def intercepted(*args:Any, **kwargs:Any) -> None:  # noqa: ANN401
            """ Wraps `print` to also log to a separate file """
            assert(self.original_print is not None)
            self.original_print(*args, **kwargs)
            if bool(args):
                print_logger.debug(args[0])

        builtins.print = intercepted

    def remove_print_capture(self) -> None:
        """ removes a previously advised builtins.print """
        match getattr(self, "original_print", None):
            case None:
                return
            case x:
                builtins.print = x
                self.original_print = None

##--|

@Mixin(PrintCapture_m)
class JGDVLogConfig(metaclass=MLSingleton):
    """ Utility class to setup [stdout, stderr, file] logging.

      Also creates a 'printer' logger, so instead of using `print`,
      tasks can notify the user using the printer,
      which also includes the notifications into the general log trace

      The Printer has a number of children, which can be controlled
      to customise verbosity.

      Standard _printer children::
        action_exec, action_group, artifact, cmd, fail, header, help, queue,
        report, skip, sleep, success, task, task_header, task_loop, task_state,
        track

    """
    ##--| classvars
    levels                : ClassVar[type[enum.IntEnum]] = API.LogLevel_e
    logger_cls            : ClassVar[type[Logger]] = JGDVLogger
    ##--| core vars
    root                  : Logger
    is_setup              : bool
    ##--| internal vars
    _initial_spec         : LoggerSpec
    _printer_initial_spec : LoggerSpec
    _registry             : dict[str, LoggerSpec]

    def __init__(self) -> None:
        # Root Logger for everything
        self.root                   = logmod.root
        self.is_setup               = False
        self._initial_spec          = INITIAL_SPEC
        self._printer_initial_spec  = PRINTER_INITIAL_SPEC
        self._registry              = {}

        self._install_logger_override()
        self._register_new_names()

        self.activate_spec(self._initial_spec)
        self.activate_spec(self._printer_initial_spec)

        logging.log(self.levels.bootstrap, "Post Log Setup") # type: ignore

    ##--| dunders
    def __repr__(self) -> str:

        return f"<{self.__class__.__name__}({len(self._registry)}) : >"

    ##--| internal
    def _install_logger_override(self) -> None:
        match self.logger_cls:
            case None:
                return
            case x if not hasattr(x, "install"):
                return
            case x:
                x.install()
                if not issubclass(logmod.getLoggerClass(), self.logger_cls):
                    msg = "Logger Class Installation Failed"
                    raise TypeError(msg, self.logger_cls)

    def _register_new_names(self) -> None:
        for name,lvl in self.levels.__members__.items(): # type: ignore
            logmod.addLevelName(lvl, name)

    def _setup_logging_extra(self, config:ChainGuard) -> None:
        """ read the doot config logging section
          setting up each entry other than stream, file, printer, and subprinters
        """
        extras = config.on_fail({}).logging.extra() # type: ignore[attr-defined]
        for key,data in extras.items():
            match LoggerSpec.build(data, name=key):
                case None:
                    logging.warning(f"Could not build LoggerSpec for {key}")
                case LoggerSpec() as spec:
                    self.activate_spec(spec)

    ##--| public
    def report(self) -> None:
        """ Utility method to inspect the state of logging. """
        printer = self.subprinter()
        for x in self._registry.values():
            printer.log(logmod.WARN, "%s : %s", x.fullname, x.level)

    def activate_spec(self, spec:LoggerSpec, *, override:bool=False) -> None:
        """ Add a spec to the registry and activate it """
        target   = spec
        fullname = spec.fullname
        logging.info("Activating Logging Spec: %s", fullname)
        self._registry[fullname] = spec
        target.apply()

    def setup(self, config:ChainGuard, *, force:bool=False) -> None:
        """ a setup that uses config values """
        if self.is_setup and not force:
            warnings.warn("Logging Is Already Set Up", stacklevel=2)

        if config is None or not bool(config):
            msg = "Config data has not been configured"
            raise ValueError(msg)

        self._initial_spec.clear()
        self._printer_initial_spec.clear()

        root_spec = LoggerSpec.build([config.on_fail({}).logging.file(), # type: ignore[attr-defined]
                                      config.on_fail({}).logging.stream(), # type: ignore[attr-defined]
                                      ],
                                     name=LoggerSpec.RootName)
        print_spec        = LoggerSpec.build(config.on_fail({}).logging.printer(), # type: ignore[attr-defined]
                                             name=API.PRINTER_NAME)

        self.activate_spec(root_spec, override=True)
        self.activate_spec(print_spec, override=True)
        self._setup_logging_extra(config)

        self.is_setup = True

    def reset(self) -> None:
        """
        Reset the config to the initial specs
        """
        self.activate_spec(self._initial_spec)
        self.activate_spec(self._printer_initial_spec)

    def set_level(self, level:int|str) -> None:
        """ Set the active logging level """
        names = logmod.getLevelNamesMapping()
        lvl   = None
        match level:
            case int():
                lvl = level
            case str() if (lvl:=names.get(level, None)) is not None:
                pass
            case _:
                msg = "Unknown Level Name"
                raise ValueError(msg, level)

        assert(lvl is not None)
        self._initial_spec.set_level(lvl)
        self._printer_initial_spec.set_level(lvl)

    def subprinter(self, *names:str, prefix:Maybe[str]=None) -> Logger:
        """ Get a subprinter of the printer logger.
          The First name needs to be a registered subprinter.
          Additional names are unconstrained
        """
        base = self._printer_initial_spec.get()
        if not bool(names) or names == (None,):
            return base

        current = base
        for name in names:
            current = current.getChild(name)
        else:
            if prefix and hasattr(current, "set_prefixes"):
                current.set_prefixes(prefix)
            return current



"""
Idenpotent Decorators, as an extendable class

Key Classes:
- DecoratorBase : Simplifies decorations to writing a _wrap_[method/fn/class] method.
- MetaDecorator : Adds metadata to callable, without changing the behaviour of it.
- DataDecorator : Stacks data onto the callable, with only one wrapping function

"""
from __future__ import annotations
import typing

# ##-- types
# isort: off
if typing.TYPE_CHECKING:
   from ._interface import Decorated, Decorator_p
   from jgdv import Maybe, Ident
   from typing import ClassVar
# isort: on
# ##-- end types

from ._interface import Signature, Decorable, DForm_e
from ._core import Decorator, MonotonicDec, IdempotentDec, MetaDec, DataDec
from .mixin import Mixin
from .proto import Proto
from .util_decorators import DoMaybe

class DecoratorAccessor_m:
    """ A mixin for building Decorator Accessors like DKeyed.
    Holds a _decoration_builder class, and helps you build it
    """

    _decoration_builder : ClassVar[type[Decorator]] = DataDec

    @classmethod
    def _build_decorator(cls, keys:list) -> Decorator_p:
        return cls._decoration_builder(keys)

    @classmethod
    def get_keys(cls, target:Decorated) -> list[Ident]:
        """ Retrieve key annotations from a Decorated"""
        dec = cls._build_decorator([])
        return dec.get_annotations(target)



#!/usr/bin/env python3
"""
A function to select an appropriate plugin by passed in name or names

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import importlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from importlib.metadata import EntryPoint
from uuid import UUID, uuid1

# ##-- end stdlib imports

from jgdv.structs.strang import CodeReference

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
   from jgdv import Maybe
   from typing import Final
   from typing import ClassVar, Any, LiteralString
   from typing import Never, Self, Literal
   from typing import TypeGuard
   from collections.abc import Iterable, Iterator, Callable, Generator
   from collections.abc import Sequence, Mapping, MutableMapping, Hashable
   from jgdv.structs.chainguard import ChainGuard

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

def plugin_selector(plugins:ChainGuard, *,
                    target:str="default",
                    fallback:Maybe[type]=None) -> Maybe[type]:
    """ Selects and loads a plugin from a chainguard,
    based on a target,
    with an available fallback constructor

    if the target is a suitable coderef, it is coerced and loaded instead,
    bypassing the plugins
    """
    logging.debug("Selecting plugin for target: %s", target)

    match target:
        case "default":
            pass
        case x:
            try:
                name = CodeReference(target)
                return name()
            except ImportError as err:
                pass
            except (AttributeError, KeyError) as err:
                pass

    match plugins:
        case [] if fallback is not None:
            return fallback
        case []:
            msg = "No Available Plugin, and no fallback constructor"
            raise ValueError(msg)
        case [EntryPoint() as l]: # if theres only one, use that
            return l.load()
        case [EntryPoint() as l, *_] if target == "default": # If the preference is the default, use the first
            return l.load()
        case [*_] as loaders: # Otherwise, use the loader that matches the preferred's name
            matching = [x for x in loaders if x.name == target]
            if bool(matching):
                return matching[0].load()
            else:
                raise ValueError("No matching plugin for target", target)
        case type():
            return x
        case _:
            msg = "Unknown type passed to plugin selector"
            raise TypeError(msg, plugins)



#!/usr/bin/env python3
"""
Adapated from typeshed

"""
# ruff: noqa
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Mapping, MutableMapping, Hashable

##--|

type ReadableBuffer    = Any
type SupportsIndex     = Any
type _FormatMapMapping = Any
type _TranslateTable   = Any
# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class String_p(Protocol):

    @staticmethod
    def maketrans[_T](x: dict[int, _T] | dict[str, _T] | dict[str | int, _T], /) -> dict[int, _T]: ...
    @staticmethod
    def maketrans(x: str, y: str, /) -> dict[int, int]: ...
    @staticmethod
    def maketrans(x: str, y: str, z: str, /) -> dict[int, int | None]: ...

    def __new__(cls, object: ReadableBuffer, encoding: str = ..., errors: str = ...) -> Self: ...
    def capitalize(self) -> str: ...
    def casefold(self) -> str: ...  # type: ignore[misc]
    def center(self, width: SupportsIndex, fillchar: str = " ", /) -> str: ...  # type: ignore[misc]
    def count(self, sub: str, start: SupportsIndex | None = ..., end: SupportsIndex | None = ..., /) -> int: ...
    def encode(self, encoding: str = "utf-8", errors: str = "strict") -> bytes: ...
    def endswith(self, suffix: str | tuple[str, ...], start: SupportsIndex | None = ..., end: SupportsIndex | None = ..., ) -> bool: ...
    def expandtabs(self, tabsize: SupportsIndex = 8) -> str: ...  # type: ignore[misc]
    def find(self, sub: str, start: SupportsIndex | None = ..., end: SupportsIndex | None = ..., /) -> int: ...
    def format(self, *args: object, **kwargs: object) -> str: ...
    def format_map(self, mapping: _FormatMapMapping, /) -> str: ...
    def index(self, sub: str, start: SupportsIndex | None = ..., end: SupportsIndex | None = ..., /) -> int: ...
    def isalnum(self) -> bool: ...
    def isalpha(self) -> bool: ...
    def isascii(self) -> bool: ...
    def isdecimal(self) -> bool: ...
    def isdigit(self) -> bool: ...
    def isidentifier(self) -> bool: ...
    def islower(self) -> bool: ...
    def isnumeric(self) -> bool: ...
    def isprintable(self) -> bool: ...
    def isspace(self) -> bool: ...
    def istitle(self) -> bool: ...
    def isupper(self) -> bool: ...
    def join(self, iterable: Iterable[str], /) -> str: ...  # type: ignore[misc]
    def ljust(self, width: SupportsIndex, fillchar: str = " ", /) -> str: ...  # type: ignore[misc]
    def lower(self) -> str: ...  # type: ignore[misc]
    def lstrip(self, chars: str | None = None, /) -> str: ...  # type: ignore[misc]
    def partition(self, sep: str, /) -> tuple[str, str, str]: ...  # type: ignore[misc]
    def replace(self: LiteralString, old: LiteralString, new: LiteralString, /, count: SupportsIndex = -1) -> LiteralString: ...
    def removeprefix(self, prefix: str, /) -> str: ...  # type: ignore[misc]
    def removesuffix(self, suffix: str, /) -> str: ...  # type: ignore[misc]
    def rfind(self, sub: str, start: SupportsIndex | None = ..., end: SupportsIndex | None = ..., /) -> int: ...
    def rindex(self, sub: str, start: SupportsIndex | None = ..., end: SupportsIndex | None = ..., /) -> int: ...
    def rjust(self, width: SupportsIndex, fillchar: str = " ", /) -> str: ...  # type: ignore[misc]
    def rpartition(self, sep: str, /) -> tuple[str, str, str]: ...  # type: ignore[misc]
    def rsplit(self, sep: str | None = None, maxsplit: SupportsIndex = -1) -> list[str]: ...  # type: ignore[misc]
    def rstrip(self, chars: str | None = None, /) -> str: ...  # type: ignore[misc]
    def split(self, sep: str | None = None, maxsplit: SupportsIndex = -1) -> list[str]: ...  # type: ignore[misc]
    def splitlines(self, keepends: bool = False) -> list[str]: ...  # type: ignore[misc]  # noqa: FBT001, FBT002
    def startswith(self, prefix: str | tuple[str, ...], start: SupportsIndex | None = ..., end: SupportsIndex | None = ..., /) -> bool: ...
    def strip(self, chars: str | None = None, /) -> str: ...  # type: ignore[misc]
    def swapcase(self) -> str: ...  # type: ignore[misc]
    def title(self) -> str: ...  # type: ignore[misc]
    def translate(self, table: _TranslateTable, /) -> str: ...
    def upper(self) -> str: ...  # type: ignore[misc]
    def zfill(self, width: SupportsIndex, /) -> str: ...  # type: ignore[misc]
    def __add__(self: LiteralString, value: LiteralString, /) -> LiteralString: ...
    def __add__(self, value: str, /) -> str: ...  # type: ignore[misc]
    # Incompatible with Sequence.__contains__
    def __contains__(self, key: str, /) -> bool: ...  # type: ignore[override]
    def __eq__(self, value: object, /) -> bool: ...
    def __ge__(self, value: str, /) -> bool: ...
    def __getitem__(self: LiteralString, key: SupportsIndex | slice, /) -> LiteralString: ...
    def __getitem__(self, key: SupportsIndex | slice, /) -> str: ...  # type: ignore[misc]
    def __gt__(self, value: str, /) -> bool: ...
    def __hash__(self) -> int: ...
    def __iter__(self) -> Iterator[str]: ...  # type: ignore[misc]
    def __le__(self, value: str, /) -> bool: ...
    def __len__(self) -> int: ...
    def __lt__(self, value: str, /) -> bool: ...
    def __mod__(self, value: Any, /) -> str: ...
    def __mul__(self, value: SupportsIndex, /) -> str: ...  # type: ignore[misc]
    def __ne__(self, value: object, /) -> bool: ...
    def __rmul__(self, value: SupportsIndex, /) -> str: ...  # type: ignore[misc]
    def __getnewargs__(self) -> tuple[str]: ...



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import logging as logmod
import warnings
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeVar, cast)
##-- end imports

import pytest
from jgdv.structs.chainguard.errors import GuardedAccessError
from jgdv.structs.chainguard._base import GuardBase
from jgdv.structs.chainguard import ChainGuard
from jgdv.structs.chainguard.proxies.failure import GuardFailureProxy

logging = logmod.root

class TestProxiedGuard:

    def test_initial(self):
        base = ChainGuard({"test": "blah"})
        proxied = base.on_fail("aweg")
        assert(isinstance(proxied, GuardFailureProxy))
        assert(isinstance(proxied.doesnt_exist, GuardFailureProxy))

    def test_proxy_on_existing_key(self):
        base = ChainGuard({"test": "blah"})
        proxied = base.on_fail("aweg")
        assert("blah" == proxied.test())

    def test_proxy_on_bad_key(self):
        base    = ChainGuard({"test": "blah"})
        proxied = base.on_fail("aweg")
        assert("aweg" == proxied.awehjo())

    def test_proxy_index_independence(self):
        base    = ChainGuard({"test": "blah"})
        base_val = base.test
        proxied = base.on_fail("aweg")
        good_key = proxied.test
        bad_key = proxied.ajojo

        assert(base._index() == ["<root>"])
        assert(proxied._index() == ["<root>"])
        assert(good_key._index() == ["<root>", "test"])
        assert(bad_key._index() == ["<root>", "ajojo"])

    def test_proxy_multi_independence(self):
        base     = ChainGuard({"test": "blah"})
        proxied  = base.on_fail("aweg")
        proxied2 = base.on_fail("jioji")
        assert(proxied is not proxied2)
        assert("aweg" == proxied.awehjo())
        assert("jioji" == proxied2.awjioq())

    def test_proxy_value_retrieval(self):
        base     = ChainGuard({"test": "blah"})
        proxied = base.on_fail("aweg").test
        assert(isinstance(proxied, GuardFailureProxy))
        assert(proxied() == "blah")

    def test_proxy_nested_value_retrieval(self):
        base     = ChainGuard({"test": { "blah": {"bloo": "final"}}})
        proxied = base.on_fail("aweg").test.blah.bloo
        assert(isinstance(proxied, GuardFailureProxy))
        assert(proxied() == "final")

    def test_proxy_none_value_use_fallback(self):
        base    = ChainGuard({"test": None})
        assert(base.test is None)
        proxied = base.on_fail("aweg").test
        assert(base.test is None)
        assert(isinstance(proxied, GuardFailureProxy))
        assert(base.test is None)
        assert(proxied._fallback == "aweg")
        assert(proxied() == "aweg")

    def test_proxy_nested_false_value_uses_fallback(self):
        base     = ChainGuard({"top": {"mid": {"bot": None}}})
        proxied = base.on_fail("aweg").top.mid.bot
        assert(isinstance(proxied, GuardFailureProxy))
        assert(proxied() is "aweg")

    def test_proxy_fallback(self):
        base     = ChainGuard({"test": { "blah": {"bloo": "final"}}})
        proxied = base.on_fail("aweg").test.blah.missing
        assert(isinstance(proxied, GuardFailureProxy))
        assert(proxied() == "aweg")

    def test_no_proxy_error(self):
        base     = ChainGuard({"test": { "blah": {"bloo": "final"}}})
        with pytest.raises(GuardedAccessError):
            base.test.blah()

    def test_proxy_early_check(self):
        base     = ChainGuard({"test": { "blah": {"bloo": "final"}}})
        proxied = base.on_fail("aweg").test
        assert(isinstance(proxied, GuardFailureProxy))

    def test_proxy_multi_use(self):
        base     = ChainGuard({"test": { "blah": {"bloo": "final", "aweg": "joijo"}}})
        proxied = base.on_fail("aweg").test.blah
        assert(proxied.bloo() == "final")
        assert(proxied.aweg() == "joijo")




#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import logging as logmod
import warnings
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple,
                    TypeVar, cast)
##-- end imports
logging = logmod.root

import pytest
from jgdv.structs.chainguard.proxies.failure import GuardFailureProxy
from jgdv.structs.chainguard.proxies.base import GuardProxy

class TestFailureProxy:

    def test_initial(self):
        proxy = GuardFailureProxy(None, fallback=2)
        assert(isinstance(proxy, GuardProxy))

    def test_attr(self):
        proxy = GuardFailureProxy(None, fallback=2)
        accessed = proxy.blah.bloo
        assert(repr(accessed) == "<GuardProxy: <root>.blah.bloo:Any>")
        assert(repr(proxy) == "<GuardProxy: <root>:Any>")

    def test_item(self):
        proxy    = GuardFailureProxy(None, fallback=2)
        accessed = proxy['blah']['bloo']
        assert(repr(accessed) == "<GuardProxy: <root>.blah.bloo:Any>")
        assert(repr(proxy) == "<GuardProxy: <root>:Any>")

    def test_multi_item(self):
        proxy    = GuardFailureProxy(None, fallback=2)
        accessed = proxy['blah', 'bloo']
        assert(repr(proxy) == "<GuardProxy: <root>:Any>")
        assert(repr(accessed) == "<GuardProxy: <root>.blah.bloo:Any>")

    def test_multi_item_expansion(self):
        proxy       = GuardFailureProxy(None, fallback=2)
        access_list = ["blah", "bloo"]
        accessed = proxy["blah", "bloo"]
        assert(repr(accessed) == "<GuardProxy: <root>.blah.bloo:Any>")

    @pytest.mark.skip(reason="FIXME")
    def test_multi_item_star_expansion(self):
        # only for py3.11+?
        proxy       = GuardFailureProxy(None, fallback=2)
        access_list = ["blah", "bloo"]
        # accessed    = proxy[*access_list]
        accessed = []
        assert(repr(accessed) == "<GuardProxy: <root>.blah.bloo:Any>")

    def test_call_get_value(self):
        proxy = GuardFailureProxy(5, fallback=2)
        assert(proxy() == 5)

    def test_call_get_None_value(self):
        proxy = GuardFailureProxy(None, fallback=None)
        assert(proxy() == None)

    def test_call_get_fallback(self):
        proxy = GuardFailureProxy(None, fallback=2)
        assert(proxy() == 2)

    def test_call_fallback_wrapper(self):
        proxy = GuardFailureProxy(None, fallback=2)
        assert(proxy(fallback_wrapper=lambda x: x*2) == 4)

    def test_wrapper(self):
        proxy = GuardFailureProxy(3, fallback=2)
        assert(proxy(wrapper=lambda x: x*2) == 6)

    def test_call_fallback_wrapper_error(self):

        def bad_wrapper(val):
            raise TypeError()

        proxy = GuardFailureProxy(None, fallback=2)
        with pytest.raises(TypeError):
            proxy(fallback_wrapper=bad_wrapper)

    def test_call_wrapper_error(self):

        def bad_wrapper(val):
            raise TypeError()

        proxy = GuardFailureProxy(3, fallback=2)
        with pytest.raises(TypeError):
            proxy(wrapper=bad_wrapper)

    def test_types(self):
        proxy = GuardFailureProxy(True, fallback=2, types=int)
        assert(proxy is not None)

    def test_types_fail(self):
        with pytest.raises(TypeError):
            GuardFailureProxy(None, fallback="blah", types=int)

    def test_proxy_inject(self):
        proxy1 = GuardFailureProxy(None, fallback=2, types=int)
        proxy2 = proxy1._inject(5)
        assert(proxy2() == 5)

    def test_proxy_value_retrieval_typecheck_fail(self):
        proxy1 = GuardFailureProxy(None, fallback=2, types=int)
        with pytest.raises(TypeError):
            proxy1._inject("blah")()

    def test_proxy_inject_index_update(self):
        proxy1 = GuardFailureProxy(None, fallback=2, types=int).blah.bloo
        proxy2 = proxy1._inject(5).awef
        assert(proxy1._index() == ["<root>", "blah", "bloo"])
        assert(proxy2._index() == ["<root>", "blah", "bloo", "awef"])
        assert(proxy2() == 5)



#!/usr/bin/env python3
"""

"""
# ruff: noqa: B011
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import warnings
from uuid import UUID, uuid1
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest

# ##-- end 3rd party imports

# ##-- 1st party imports
from .. import _interface as API
from .._core import (
    Decorator,
    IdempotentDec,
    MonotonicDec,
)

from jgdv.decorators.proto import Proto
# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

logging = logmod.root

##-- protocols

class AbsClass(abc.ABC):
    """ An Abstract class with an explicit abstract method """

    @abc.abstractmethod
    def blah(self):
        pass

@runtime_checkable
class AbsProto_p(Protocol):
    """ A Protocol with an explicit abstract method """

    def blah(self) -> None: ...

    @abc.abstractmethod
    def other(self):
        pass

@runtime_checkable
class RawProto_p(Protocol):
    """ A Protocol with no additional anntoations """

    def blah(self): ...

    def aweg(self): ...

##-- end protocols

##-- implementations

class GoodAbsClass(AbsClass):

    def blah(self):
        return 2

class BadAbsClass(AbsClass):
    pass

##--|

class GoodInheritAbsProto(AbsProto_p):

    def blah(self) -> None:
        pass

    def other(self):
        pass

class BadInheritAbsProto(AbsProto_p):
    pass

##--|

class GoodInheritRawProto(RawProto_p):

    def blah(self):
        return 10

    def aweg(self):
        return 10

class BadInheritRawProto(RawProto_p):

    def aweg(self):
        return 10

##--|

class GoodStructuralRawProto:

    def blah(self):
        return 10

    def aweg(self):
        return 10

class BadStructuralRawProto:

    def aweg(self):
        return 10

##-- end implementations

class TestProtoDecorator:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_proto_no_check_no_error(self):

        @Proto(RawProto_p, check=False, mod_mro=True)
        class Example:
            """ doesn't implement the protocol, but is annotated with it """
            val : ClassVar[int] = 25

            def bloo(self):
                return 10
            
            def aweg(self):
                return blah

        obj = Example()
        assert(isinstance(Example, RawProto_p))
        assert(RawProto_p in Example.mro())
        assert(obj.bloo() == 10)
        assert(Example.val == 25)
        match Proto.get(Example):
            case [x] if x is RawProto_p:
                assert(True)
            case x:
                 assert(False), x

    def test_proto_no_check(self):

        @Proto(RawProto_p, check=False)
        class Example:
            val : ClassVar[int] = 25

            def blah(self):
                return 10

            def aweg(self):
                return 10

        obj = Example()
        assert(isinstance(Example, RawProto_p))
        assert(obj.blah() == 10)
        assert(obj.aweg() == 10)
        assert(Example.val == 25)
        match Proto.get(Example):
            case [x] if x is RawProto_p:
                assert(True)
            case x:
                 assert(False), x

    def test_proto_check_success(self):

        @Proto(RawProto_p, check=True)
        class Example:
            val : ClassVar[int] = 25

            def bloo(self):
                return 10

            def blah(self):
                return

            def aweg(self):
                return

        obj = Example()
        assert(isinstance(Example, RawProto_p))
        assert(obj.bloo() == 10)
        assert(Example.val == 25)
        match Proto.get(Example):
            case [x] if x is RawProto_p:
                assert(True)
            case x:
                 assert(False), x

    def test_proto_check_fail(self):

        with pytest.raises(NotImplementedError):
            @Proto(RawProto_p, check=True)
            class Example:
                val : ClassVar[int] = 25

                def bloo(self):
                    return 10

    def test_proto_on_type_alias(self):
        type proto_alias = RawProto_p

        @Proto(proto_alias, check=True)
        class Example:
            val : ClassVar[int] = 25

            def bloo(self):
                return 10

            def blah(self):
                return

            def aweg(self):
                return

        assert(True)

    def test_proto_on_protocol_fails(self):

        with pytest.raises(TypeError) as ctx:
            @Proto(RawProto_p)
            class Example(Protocol):
                pass

        match ctx.value:
            case TypeError(args=["Don't use @Proto to combine protocols, use normal inheritance", *_]):
                assert(True)
            case x:
                 assert(False), x


    def test_proto_annotations(self):
        assert(isinstance(GoodStructuralRawProto, RawProto_p))
        @Proto(RawProto_p)
        class Sub(GoodStructuralRawProto):
            pass

        match Proto.get(Sub):
            case [x] if x == RawProto_p:
                assert(True)
            case x:
                assert(False), x


    def test_proto_annotations_dont_travel_to_superclass(self):
        assert(isinstance(GoodStructuralRawProto, RawProto_p))
        @Proto(RawProto_p)
        class Sub(GoodStructuralRawProto):
            pass

        match Proto.get(GoodStructuralRawProto):
            case []:
                assert(True)
            case x:
                assert(False), x
class TestCheckProtocolClass:

    @pytest.fixture(scope="function")
    def proto(self):
        return Proto()

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_abs_class_success(self, proto):
        assert(issubclass(GoodAbsClass, AbsClass))
        proto.validate_protocols(GoodAbsClass)
        assert(True)

    def test_abs_class_fail(self, proto):
        assert(issubclass(BadAbsClass, AbsClass))
        with pytest.raises(NotImplementedError):
            proto.validate_protocols(BadAbsClass)

    def test_abs_proto_class_success(self, proto):
        assert(isinstance(GoodInheritAbsProto, AbsProto_p))
        proto.validate_protocols(GoodInheritAbsProto)
        assert(True)

    def test_abs_proto_class_fail(self, proto):
        assert(issubclass(BadInheritAbsProto, AbsProto_p))
        with pytest.raises(NotImplementedError):
            proto.validate_protocols(BadInheritAbsProto)
            assert(True)

    def test_raw_proto_class_success(self, proto):
        assert(isinstance(GoodInheritRawProto, RawProto_p))
        proto.validate_protocols(GoodInheritRawProto)
        assert(True)

    def test_raw_proto_class_fail(self, proto):
        assert(issubclass(BadInheritRawProto, RawProto_p))
        with pytest.raises(NotImplementedError):
            proto.validate_protocols(BadInheritRawProto)

    def test_raw_structural_proto_class_success(self):
        assert(isinstance(GoodStructuralRawProto, RawProto_p))
        assert(not RawProto_p in GoodStructuralRawProto.mro())

        @Proto(RawProto_p)
        class Sub(GoodStructuralRawProto):
            pass

        match Proto.get(Sub):
            case [x] if x == RawProto_p:
                assert(True)
            case x:
                assert(False), x

        match Proto.get(GoodStructuralRawProto):
            case []:
                assert(True)
            case x:
                assert(False), x

    def test_raw_structural_proto_class_fail(self):
        assert(not issubclass(BadStructuralRawProto, RawProto_p))
        assert(not isinstance(BadStructuralRawProto, RawProto_p))
        assert(not RawProto_p in BadStructuralRawProto.mro())
        with pytest.raises(NotImplementedError):
            Proto.validate_protocols(BadStructuralRawProto, RawProto_p)
            assert(True)


    @pytest.mark.parametrize("cls", [GoodStructuralRawProto])
    def test_get_protocols_empty(self, cls):
        match Proto.get(cls):
            case []:
                assert(True)
            case x:
                assert(False), x


    @pytest.mark.parametrize("cls", [GoodAbsClass, GoodInheritAbsProto, GoodInheritRawProto])
    def test_get_protocols(self, cls):
        match Proto.get(cls):
            case list() as xs if bool(xs):
                 assert(True)
            case x:
                 assert(False), (x, cls)

class TestGenericProtocolChecking:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_unparam_generic_proto_successs(self):

        @runtime_checkable
        class GenericProto_p[T](Protocol):
            pass

        @Proto(GenericProto_p)
        class Impl:
            pass

        assert(True)

    def test_param_generic_proto_success(self):

        @runtime_checkable
        class GenericProto_p[T](Protocol):
            pass

        @Proto(GenericProto_p[int])
        class Impl:
            pass

        assert(True)

    def test_aliased_param_generic_proto_success(self):

        @runtime_checkable
        class GenericProto_p[T](Protocol):
            pass

        type IntGenericProto_p = GenericProto_p[int]

        @Proto(IntGenericProto_p)
        class Impl:
            pass

        assert(True)

    def test_unparam_generic_proto_fail(self):

        @runtime_checkable
        class GenericProto_p[T](Protocol):

            def blah(self): ...

        with pytest.raises(NotImplementedError):

            @Proto(GenericProto_p)
            class Impl:
                pass

    def test_param_generic_proto_fail(self):

        @runtime_checkable
        class GenericProto_p[T](Protocol):

            def blah(self): ...

        with pytest.raises(NotImplementedError):

            @Proto(GenericProto_p[int])
            class Impl:
                pass

    def test_alias_param_generic_proto_fail(self):

        @runtime_checkable
        class GenericProto_p[T](Protocol):

            def blah(self): ...

        type IntGenProto_p = GenericProto_p[int]
        with pytest.raises(NotImplementedError):

            @Proto(IntGenProto_p)
            class Impl:
                pass



#!/usr/bin/env python3
"""

"""
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
import warnings

import pytest

from .. import _interface as API # noqa: N812
from jgdv.structs.dkey import DKey
from jgdv.structs.strang import Strang, StrangError
from jgdv.structs.locator import JGDVLocator, Location

logging = logmod.root

class TestLocation:

    def test_sanity(self):
        assert(True is not False)

    def test_simple_dir(self):
        loc = Location("dir::>test/path")
        assert(loc is not None)
        assert(isinstance(loc, Strang))
        assert(isinstance(loc, Location))
        assert(isinstance(loc, API.Location_p))
        assert(isinstance(loc, str))

    def test_simple_file(self):
        loc = Location("file::>test/path.py")
        assert(loc is not None)
        assert(isinstance(loc, Strang))
        assert(isinstance(loc, Location))
        assert(isinstance(loc, str))
        assert(Location.Marks.file in loc)


    def test_simple_abstract(self):
        target = "test/*/path.py"
        loc = Location("file::>test/*/path.py")
        assert(loc is not None)
        assert(isinstance(loc, Location))
        assert(loc[1,:] == target)
        assert(Location.Marks.file in loc)
        assert(Location.Marks.abstract in loc)

    def test_file_stem(self):
        loc = Location("file::>test/path.py")
        assert(loc.path.stem == "path")
        assert(loc.stem == "path")

    def test_file_ext(self):
        loc = Location("file::>test/path.py")
        assert(loc.ext() == ".py")

    def test_file_multi_ext(self):
        loc = Location("file::>test/path.py.bl.gz")
        assert(loc.ext() == ".py.bl.gz")

    def test_multi_ext_last(self):
        loc = Location("file::>test/path.py.bl.gz")
        assert(loc.ext(last=True) == ".gz")
        assert(loc.ext(last=False) == ".py.bl.gz")

    def test_bad_form_fail(self):
        # with pytest.raises(StrangError):
        #     Location("bad::test/path.py")
        val = Location("bad::test/path.py")

    def test_file_with_metadata(self):
        loc = Location("file.clean::>test/path.py")
        assert(isinstance(loc, Location))
        assert(Location.Marks.file in loc.data.meta)
        assert(Location.Marks.clean in loc.data.meta)
        assert(Location.Marks.abstract not in loc.data.meta)

    def test_glob_path(self):
        loc = Location("file::>test/*/path.py")
        assert(isinstance(loc, Location))
        assert(Location.Marks.abstract in loc.data.meta)
        assert(not loc.is_concrete())
        assert(loc.get(1,1) == loc.Wild.glob)

    def test_rec_glob_path(self):
        loc = Location("file::>test/**/path.py")
        assert(isinstance(loc, Location))
        assert(Location.Marks.abstract in loc.data.meta)
        assert(not loc.is_concrete())
        assert(loc.get(1,1) == loc.Wild.rec_glob)

    def test_select_path(self):
        loc = Location("file::>test/?/path.py")
        assert(isinstance(loc, Location))
        assert(Location.Marks.abstract in loc.data.meta)
        assert(not loc.is_concrete())
        assert(loc.get(1,1) == loc.Wild.select)

    def test_glob_stem(self):
        loc = Location("file::>test/blah/*ing.py")
        assert(isinstance(loc, Location))
        assert(Location.Marks.abstract in loc.data.meta)
        assert(not loc.is_concrete())
        assert(loc.stem == (loc.Wild.glob, "*ing"))

    def test_select_stem(self):
        loc = Location("file::>test/blah/?ing.py")
        assert(isinstance(loc, Location))
        assert(Location.Marks.abstract in loc.data.meta)
        assert(not loc.is_concrete())
        assert(loc.stem == (loc.Wild.select, "?ing"))

    def test_earlycwd_expansion(self):
        loc = Location("file/earlycwd::>a/b/c.py")
        assert(loc[1,:] == "a/b/c.py")
        assert(loc.path == pl.Path("a/b/c.py"))

    def test_earlycwd_expansion_uses_initial_cwd(self):
        loc       = Location("file.earlycwd::>a/b/c.py")
        orig_cwd  = pl.Path.cwd()
        sub_cwd   = [x for x in orig_cwd.iterdir() if not x.is_file()][0]
        with JGDVLocator(sub_cwd) as loclookup:
            assert(pl.Path.cwd() != orig_cwd)
            assert(pl.Path.cwd() == sub_cwd)
            expanded = loclookup.normalize(loc)
            assert(expanded.is_absolute())
            assert(not expanded.is_relative_to(sub_cwd))

class TestLocation_Definite:

    def test_definite(self):
        basic = Location(pl.Path("a/b/c"))
        assert(basic.is_concrete())

    def test_definite_contained_in_indefinite(self):
        definite = Location("a/b/c.py")
        indef    = Location("a/b/*.py")
        assert(Location.Marks.abstract in indef)
        assert(Location.Marks.abstract not in definite)
        assert(definite in indef)

class TestLocation_Indefinite:

    def test_indef_stem_not_concrete(self):
        basic = Location(pl.Path("a/b/*.py"))
        assert(not basic.is_concrete())

    def test_indef_ext_not_concrete(self):
        basic = Location(pl.Path("a/b/c.*"))
        assert(basic.ext() == (Location.Wild.glob, ".*"))
        assert(not basic.is_concrete())

    def test_indef_path_not_concrete(self):
        basic = Location(pl.Path("a/*/c.py"))
        assert(basic.Wild.glob in basic.body)
        assert(not basic.is_concrete())

    def test_indef_path_not_concrete_2(self):
        basic = Location(pl.Path("a/**/c.py"))
        assert(basic.Wild.rec_glob in basic.body)
        assert(not basic.is_concrete())

    def test_indef_suffix_contains_definite(self):
        definite = Location(pl.Path("a/b/c.py"))
        indef    = Location(pl.Path("a/b/c.*"))
        assert(indef.ext() == (indef.Wild.glob, ".*"))
        assert(definite in indef)

    def test_indef_suffix_contain_fail(self):
        definite = Location(pl.Path("a/b/d.py"))
        indef    = Location(pl.Path("a/b/c.*"))
        assert(definite not in indef)

    def test_indef_path_contains(self):
        definite = Location(pl.Path("a/b/c.py"))
        indef    = Location(pl.Path("a/*/c.py"))
        assert(definite in indef)

    def test_indef_path_contain_fail(self):
        definite = Location(pl.Path("a/b/d/c.py"))
        indef    = Location(pl.Path("a/*/e/c.py"))
        assert(indef not in definite)

    def test_indef_stem_contains_definite(self):
        definite = Location(pl.Path("a/b/c.py"))
        indef    = Location(pl.Path("a/b/*.py"))
        assert(definite in indef)

    def test_indef_stem_contains_fail(self):
        definite = Location(pl.Path("a/b/c.bib"))
        indef    = Location(pl.Path("a/b/*.py"))
        assert(definite not in indef)

    def test_indef_recursive_contains(self):
        definite = Location(pl.Path("a/b/d/c.py"))
        indef    = Location(pl.Path("a/**/c.py"))
        assert(definite in indef)

    def test_indef_recursive_contain_fail(self):
        definite = Location(pl.Path("b/b/c.py"))
        indef    = Location(pl.Path("a/**/c.py"))
        assert(definite not in indef)

    def test_indef_multi_recursive_contains(self):
        definite = Location(pl.Path("a/b/d/e/f/c.py"))
        indef    = Location(pl.Path("a/**/c.py"))
        assert(definite in indef)

    def test_indef_root_recursive_contains(self):
        definite = Location(pl.Path("a/b/d/e/f/c.py"))
        indef    = Location(pl.Path("**/c.py"))
        assert(definite in indef)

    def test_indef_multi_component_contains(self):
        definite = Location(pl.Path("a/b/d/e/f/c.py"))
        indef    = Location(pl.Path("**/*.*"))
        assert(definite in indef)



#!/usr/bin/env python3
"""

"""
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
import warnings

import pytest

logging = logmod.root

from jgdv.files.tags.tag_file import TagFile

class TestTagFile:

    def test_initial(self):
        obj = TagFile()
        assert(obj is not None)

    @pytest.mark.parametrize("tag,exp", [("tag", "tag"), ("a tag", "a_tag"), ("A Tag", "A_Tag")])
    def test_norm_tag(self, tag, exp):
        obj = TagFile()
        normed = obj.norm_tag(tag)
        assert(normed == exp)

    def test_empty(self):
        obj = TagFile()
        assert(not bool(obj))

    def test_non_empty(self):
        obj = TagFile(counts={"blah": 1})
        assert(bool(obj))

    def test_len(self):
        obj = TagFile(counts={"blah": 1})
        assert(len(obj) == 1)

    def test_len_2(self):
        obj = TagFile(counts={"blah": 1, "bloo": 3})
        assert(len(obj) == 2)

    def test_contains(self):
        obj = TagFile(counts={"blah": 1})
        assert("blah" in obj)

    def test_contains_fail(self):
        obj = TagFile(counts={"blah": 1})
        assert("bloo" not in obj)

    def test_contains_norms(self):
        obj = TagFile(counts={"a blah": 1, "a_bloo":5})
        assert("a blah" in obj)
        assert("a_blah" in obj)
        assert("a_bloo" in obj)
        assert("a bloo" in obj)

    def test_update_str(self):
        obj = TagFile()
        assert(not bool(obj))
        assert("bloo" not in obj)
        obj.update("bloo")
        assert(bool(obj))
        assert("bloo" in obj)

    def test_update_list(self):
        obj = TagFile()
        assert(not bool(obj))
        assert("bloo" not in obj)
        obj.update(["bloo", "blee"])
        assert(bool(obj))
        assert("bloo" in obj)

    def test_update_set(self):
        obj = TagFile()
        assert(not bool(obj))
        obj.update({"bloo", "blah", "blee"})
        assert(bool(obj))
        assert("bloo" in obj)
        assert("blah" in obj)
        assert("blee" in obj)

    def test_update_str_multi(self):
        obj = TagFile()
        assert(not bool(obj))
        assert("bloo" not in obj)
        obj.update("bloo", "blah")
        assert(bool(obj))
        assert("bloo" in obj)
        assert("blah" in obj)

    def test_update_dict(self):
        obj = TagFile()
        assert(not bool(obj))
        obj.update({"bloo":1, "blah":3, "blee":5})
        assert(bool(obj))
        assert("bloo" in obj)
        assert("blah" in obj)
        assert("blee" in obj)

    def test_update_tagfile(self):
        obj = TagFile()
        obj2 = TagFile(counts={"blah":1, "bloo":1, "blee":1})
        assert(not bool(obj))
        obj.update(obj2)
        assert(bool(obj))
        assert("bloo" in obj)
        assert("blah" in obj)
        assert("blee" in obj)

    def test_to_set(self):
        obj = TagFile(counts={"blah":1, "bloo":1, "blee":1})
        as_set = obj.to_set()
        assert(isinstance(as_set, set))
        assert(len(as_set) == 3)

    def test_get_count(self):
        obj = TagFile(counts={"blah":1, "bloo":5, "blee":1})
        assert(obj.get_count("blah") == 1)

    def test_get_count_2(self):
        obj = TagFile(counts={"blah":1, "bloo":5, "blee":1})
        assert(obj.get_count("bloo") == 5)

    def test_get_count_missing(self):
        obj = TagFile(counts={"blah":1, "bloo":5, "blee":1})
        assert(obj.get_count("aweg") == 0)

    def test_count_inc(self):
        obj = TagFile(counts={"blah":1, "bloo":5, "blee":1})
        assert(obj.get_count("bloo") == 5)
        obj.update("bloo")
        assert(obj.get_count("bloo") == 6)

    def test_str(self):
        obj = TagFile(counts={"blah":1, "bloo":5, "blee":1, "aweg": 0})
        assert(str(obj) == "\n".join(["blah : 1", "blee : 1", "bloo : 5"]))




"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202

# Imports
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
import warnings

import pytest
from ..logger import JGDVLogger

# Logging:
logging = logmod.root

# Type Aliases:

# Vars:

# Body:

class TestJGDVLogger:

    @pytest.fixture(scope="function")
    def install(self):
        orig = logmod.getLoggerClass()
        JGDVLogger.install()
        yield
        logmod.setLoggerClass(orig)

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        logger = JGDVLogger("basic")
        assert(isinstance(logger, logmod.getLoggerClass()))

    def test_defaults(self):
        logger = JGDVLogger("basic")
        assert(hasattr(logger, "info"))
        assert(hasattr(logger, "debug"))
        assert(hasattr(logger, "warning"))
        assert(hasattr(logger, "error"))
        assert(hasattr(logger, "critical"))

    def test_non_default_fail(self):
        logger = JGDVLogger("basic")
        with pytest.raises(AttributeError):
            getattr(logger, "blah")

        with pytest.raises(AttributeError):
            logger.blah

    def test_non_default_success(self, install, caplog):
        with caplog.at_level(logmod.DEBUG):
            logger = logmod.getLogger("basic")
            assert(isinstance(logger, JGDVLogger))
            assert(callable(logger.trace))
            logger.trace("blah")

        assert("blah" in caplog.messages)

    def test_non_default_getitem(self, install, caplog):
        with caplog.at_level(logmod.DEBUG):
            logger = logmod.getLogger("basic")
            assert(isinstance(logger, JGDVLogger))
            assert(callable(logger['trace']))
            logger['trace']("blah")

        assert("blah" in caplog.messages)

    def test_prefix(self, install, caplog):
        with caplog.at_level(logmod.DEBUG):
            logger = logmod.getLogger("basic")
            logger.trace("blah")
            logger.prefix("> ").trace("bloo")
            logger.prefix("> ").trace("qqqq")
            logger.trace("aweg")

        assert("blah" in caplog.messages)
        assert("> bloo" in caplog.messages)
        assert("> qqqq" in caplog.messages)
        assert("aweg" in caplog.messages)

    def test_callable_prefix(self, install, caplog):
        count = 0

        def the_prefix():
            nonlocal count
            count += 1
            return f"({count}) "

        with caplog.at_level(logmod.DEBUG):
            logger = logmod.getLogger("basic")
            logger.trace("blah")
            logger.prefix(the_prefix).trace("bloo")
            logger.prefix(the_prefix).trace("aweg")
            logger.trace("aweg")

        assert("blah" in caplog.messages)
        assert("(1) bloo" in caplog.messages)
        assert("(2) aweg" in caplog.messages)
        assert("aweg" in caplog.messages)



#/usr/bin/env python3
"""

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
from collections import ChainMap
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from re import Pattern
from uuid import UUID, uuid1
from weakref import ref
import tomllib
# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv import Proto, Mixin

from ._base import GuardBase
from .errors import GuardedAccessError
from .mixins.access_m import TomlAccess_m
from .mixins.loader_m import TomlLoader_m
from .mixins.proxy_m import GuardProxyEntry_m
from .mixins.reporter_m import DefaultedReporter_m
from .mixins.writer_m import TomlWriter_m

from ._interface import TomlTypes, ChainGuard_p
# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Proto(ChainGuard_p)
@Mixin(GuardProxyEntry_m, TomlLoader_m, TomlWriter_m, silent=True)
@Mixin(TomlAccess_m, DefaultedReporter_m, silent=True)
class ChainGuard(GuardBase):
    """ The Final ChainGuard class.

    Takes the GuardBase object, and mixes in extra capabilities.

    """
    @classmethod
    def merge(cls, *guards:Self, dfs:Maybe[Callable]=None, index:Maybe[str]=None, shadow:bool=False) -> Self:  # noqa: ARG003
        """
        Given an ordered list of guards and dicts, convert them to dicts,
        update an empty dict with each,
        then wrap that combined dict in a ChainGuard

        *NOTE*: classmethod, not instance. search order is same as arg order.
        So merge(a, b, c) will retrive from c only if a, then b, don't have the key

        # TODO if given a dfs callable, use it to merge more intelligently
        """
        curr_keys : set = set()
        # Check for conflicts:
        for data in guards:
            new_keys = set(data.keys())
            if bool(curr_keys & new_keys) and not shadow:
                msg = "Key Conflict:"
                raise KeyError(msg, curr_keys & new_keys)
            curr_keys |= new_keys

        # Build a TG from a chainmap
        return ChainGuard.from_dict(ChainMap(*(dict(x) for x in guards))) # type: ignore

    def remove_prefix(self, prefix:str) -> ChainGuard:
        """ Try to remove a prefix from loaded data
          eg: ChainGuard(tools.ChainGuard.data..).remove_prefix("tools.ChainGuard")
          -> ChainGuard(data..)
        """
        match prefix:
            case None:
                return self
            case str():
                logging.debug("Removing prefix from data: %s", prefix)
                try:
                    attempt = self
                    for x in prefix.split("."):
                        attempt = attempt[x]
                    else:
                        return ChainGuard(attempt)
                except GuardedAccessError:
                    return self



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import logging as logmod
import warnings
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple,
                    TypeVar, cast)
##-- end imports

import typing
import pytest
from jgdv.structs.chainguard.errors import GuardedAccessError
from jgdv.structs.chainguard import ChainGuard

logging = logmod.root

class TestBaseGuard:

    def test_initial(self):
        basic = ChainGuard({"test": "blah"})
        assert(basic is not None)

    def test_is_mapping(self):
        basic = ChainGuard({"test": "blah"})
        assert(isinstance(basic, typing.Mapping))


    def test_is_dict(self):
        basic = ChainGuard({"test": "blah"})
        assert(isinstance(basic, dict))


    def test_match_as_dict(self):
        match ChainGuard({"test": "blah"}):
            case dict():
                assert(True)
            case x:
                 assert(False), x


    def test_basic_access(self):
        basic = ChainGuard({"test": "blah"})
        assert(basic.test == "blah")

    def test_basic_item_access(self):
        basic = ChainGuard({"test": "blah"})
        assert(basic['test'] == "blah")

    def test_multi_item_access(self):
        basic = ChainGuard({"test": {"blah": "bloo"}})
        assert(basic['test', "blah"] ==  "bloo")

    def test_basic_access_error(self):
        basic = ChainGuard({"test": "blah"})
        with pytest.raises(GuardedAccessError):
            basic.none_existing

    def test_item_access_error(self):
        basic = ChainGuard({"test": "blah"})
        with pytest.raises(GuardedAccessError):
            basic['non_existing']

    def test_dot_access(self):
        basic = ChainGuard({"test": "blah"})
        assert(basic.test == "blah")

    def test_index(self):
        basic = ChainGuard({"test": "blah"})
        assert(basic._index() == ["<root>"])

    def test_index_independence(self):
        basic = ChainGuard({"test": "blah"})
        assert(basic._index() == ["<root>"])
        basic.test
        assert(basic._index() == ["<root>"])

    def test_nested_access(self):
        basic = ChainGuard({"test": {"blah": 2}})
        assert(basic.test.blah == 2)

    def test_repr(self):
        basic = ChainGuard({"test": {"blah": 2}, "bloo": 2})
        assert(repr(basic) == "<ChainGuard:['test', 'bloo']>")

    def test_immutable(self):
        basic = ChainGuard({"test": {"blah": 2}, "bloo": 2})
        with pytest.raises(TypeError):
            basic.test = 5

    def test_uncallable(self):
        basic = ChainGuard({"test": {"blah": 2}, "bloo": 2})
        with pytest.raises(GuardedAccessError):
            basic()

    def test_iter(self):
        basic = ChainGuard({"test": {"blah": 2}, "bloo": 2})
        vals = list(basic)
        assert(vals == ["test", "bloo"])

    def test_contains(self):
        basic = ChainGuard({"test": {"blah": 2}, "bloo": 2})
        assert("test" in basic)

    def test_contains_fail(self):
        basic = ChainGuard({"test": {"blah": 2}, "bloo": 2})
        assert("blah" not in basic)

    def test_get(self):
        basic = ChainGuard({"test": {"blah": 2}, "bloo": 2})
        assert(basic.get("bloo") == 2)

    def test_get_default(self):
        basic = ChainGuard({"test": {"blah": 2}, "bloo": 2})
        assert(basic.get("blah") is None)

    def test_get_default_value(self):
        basic = ChainGuard({"test": {"blah": 2}, "bloo": 2})
        assert(basic.get("blah", 5) == 5)

    def test_keys(self):
        basic = ChainGuard({"test": {"blah": 2}, "bloo": 2})
        assert(list(basic.keys()) == ["test", "bloo"])

    def test_items(self):
        basic = ChainGuard({"test": {"blah": 2}, "bloo": 2})
        assert(list(basic.items()) == [("test", {"blah": 2}), ("bloo", 2)])

    def test_values(self):
        basic = ChainGuard({"test": {"blah": 2}, "bloo": 2})
        assert(list(basic.values()) == [{"blah": 2}, 2])

    def test_list_access(self):
        basic = ChainGuard({"test": {"blah": [1,2,3]}, "bloo": ["a","b","c"]})
        assert(basic.test.blah == [1,2,3])
        assert(basic.bloo == ["a","b","c"])

    def test_contains(self):
        basic = ChainGuard({"test": {"blah": [1,2,3]}, "bloo": ["a","b","c"]})
        assert("test" in basic)

    def test_contains_false(self):
        basic = ChainGuard({"test": {"blah": [1,2,3]}, "bloo": ["a","b","c"]})
        assert("doesntexist" not in basic)

    def test_contains_nested_but_doesnt_recurse(self):
        basic = ChainGuard({"test": {"blah": [1,2,3]}, "bloo": ["a","b","c"]})
        assert("blah" not in basic)

    def test_contains_nested(self):
        basic = ChainGuard({"test": {"blah": [1,2,3]}, "bloo": ["a","b","c"]})
        assert("blah" in basic.test)

    def test_contains_nested_false(self):
        basic = ChainGuard({"test": {"blah": [1,2,3]}, "bloo": ["a","b","c"]})
        assert("doesntexist" not in basic.test)

class TestLoaderGuard:

    @pytest.mark.skip(reason="not implemented")
    def test_initial_load(self):
        # TODO
        raise

class TestGuardMerge:

    def test_initial(self):
        simple = ChainGuard.merge({"a":2}, {"b": 5})
        assert(isinstance(simple, ChainGuard))
        assert(simple._table() == {"a": 2, "b": 5})

    def test_merge_conflict(self):
        with pytest.raises(KeyError):
            ChainGuard.merge({"a":2}, {"a": 5})

    def test_merge_with_shadowing(self):
        basic = ChainGuard.merge({"a":2}, {"a": 5, "b": 5}, shadow=True)
        assert(dict(basic) == {"a":2, "b": 5})

    def test_merge_guards(self):
        first  = ChainGuard({"a":2})
        second = ChainGuard({"a": 5, "b": 5})

        merged = ChainGuard.merge(first ,second, shadow=True)
        assert(dict(merged) == {"a":2, "b": 5})

    def test_dict_updated_with_chainguard(self):
        the_dict = {}
        cg = ChainGuard({"a": 2, "b": 3, "c": {"d": "test" }})
        assert(not bool(the_dict))
        the_dict.update(cg)
        assert(bool(the_dict))
        assert("a" in the_dict)
        assert(the_dict["c"]["d"] == "test")

class TestFailAccess:

    def test_basic(self):
        obj = ChainGuard({})
        assert(obj is not  None)

    def test_basic_fail(self):
        obj = ChainGuard({})
        result = obj.on_fail(5).nothing()
        assert(result == 5)

    def test_fail_access_dict(self):
        obj = ChainGuard({"nothing": {}})
        result = obj.on_fail({}).nothing['blah']()
        assert(isinstance(result, dict))

    def test_fail_access_list(self):
        obj = ChainGuard({"nothing": []})
        result = obj.on_fail([]).nothing[1]()
        assert(isinstance(result, list))

    def test_fail_access_type_mismatch(self):
        obj = ChainGuard({"nothing": {}})
        result = obj.on_fail({}).nothing[1]()
        assert(isinstance(result, dict))

    def test_fail_return_none(self):
        obj = ChainGuard({"nothing": {}})
        result = obj.on_fail(None).nothing.blah.bloo()
        assert(result is None)


    def test_success_return_val(self):
        obj = ChainGuard({"nothing": {"blah": {"bloo": 10}}})
        result = obj.on_fail(None).nothing.blah.bloo()
        assert(result == 10)



#!/usr/bin/env python3
"""
The Interface for Strang.

Strang Enums:
- StrangMarkAbstract_e

Describes the internal data structs:
- Sec_d : A Single section spec
- Sections_d : Collects the sec_d's. ClassVar
- Strang_d : Instance data of a strang beyond the normal str's

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
# ##-- end stdlib imports

from jgdv._abstract.str_proto import String_p

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from collections.abc import Sized
from collections import UserString
from types import UnionType

if TYPE_CHECKING:
    from jgdv._abstract.pre_processable import PreProcessor_p
    from jgdv import Maybe, Rx
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard, SupportsIndex
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable
    import string
    from enum import Enum

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--| Vars
FMT_PATTERN   : Final[Rx]                    = re.compile("^(h?)(t?)(p?)")
TYPE_RE       : Final[Rx]                    = re.compile(r"<(.+?)(?::(.+?))?>") # TODO name the groups
TYPE_ITER_RE  : Final[Rx]                    = re.compile(r"(<)(.+?)(?::(.+?))?(>)")
MARK_RE       : Final[Rx]                    = re.compile(r"(\$.+?\$)")
MARK_ITER_RE  : Final[Rx]                    = re.compile(r"(\$)(.+?)(\$)")
ARGS_RE       : Final[Rx]                    = re.compile(r"\[(.+?)\]$")
ARGS_CHARS    : Final[tuple[str, str, str]]  = "[", ",", "]"
CASE_DEFAULT  : Final[str]                   = "."
END_DEFAULT   : Final[str]                   = "::"
INST_K        : Final[str]                   = "instanced"
GEN_K         : Final[str]                   = "gen_uuid"
STRGET        : Final[Callable]              = str.__getitem__
STRCON        : Final[Callable]              = str.__contains__
UUID_WORD     : Final[str]                   = "<uuid>"

SEC_END_MSG   : Final[str]                   = "Only the last section has no end marker"

##--| Enums

class StrangMarkAbstract_e(enum.StrEnum):

    @classmethod
    def default(cls) -> Maybe:
        return None

    @classmethod
    def implicit(cls) -> set:
        return set()

    @classmethod
    def skip(cls) -> Maybe:
        return None

    @classmethod
    def idempotent(cls) -> set[str]:
        return set()

class DefaultHeadMarks_e(StrangMarkAbstract_e):
    """ Markers used in a Strang's head """
    basic = "$basic$"

class DefaultBodyMarks_e(StrangMarkAbstract_e):
    """ Markers Used in a base Strang's body """

    head    = "$head$"
    gen     = "$gen$"
    empty   = ""
    hide    = "_"
    extend  = "+"

    @classmethod
    def default(cls) -> str:
        return cls.head

    @classmethod
    def implicit(cls) -> set[str]:
        return {cls.hide, cls.empty}

    @classmethod
    def skip(cls) -> Maybe[str]:
        return cls.empty

    @classmethod
    def idempotent(cls) -> set[str]:
        return {cls.head, cls.gen}

class CodeRefHeadMarks_e(StrangMarkAbstract_e):
    """ Available Group values of CodeRef strang's """
    module  = "module"
    cls     = "cls"
    value   = "value"
    fn      = "fn"

    val     = "value"

    @classmethod
    def default(cls) -> str:
        return cls.fn

    @classmethod
    def idempotent(cls) -> set[str]:
        return {cls.module, cls.cls, cls.value, cls.fn}


##--|
type FullSlice     = slice[None, None, None]
type MSlice        = slice[Maybe[int], Maybe[int], Maybe[int]]
type SectionIndex  = str|int
type WordIndex     = tuple[SectionIndex, int]
type MarkIndex     = tuple[SectionIndex, StrangMarkAbstract_e]
type FindSlice     = str|StrangMarkAbstract_e|WordIndex|MarkIndex
type ItemIndex     = SectionIndex | FullSlice | MSlice | tuple[ItemIndex, ...]
type PushVal       = Maybe[str | StrangMarkAbstract_e | UUID]

HEAD_TYPES    = str|UUID|DefaultBodyMarks_e
BODY_TYPES    = str|UUID|DefaultBodyMarks_e
##--| Data

class Sec_d:
    """ Data of a named Strang section

    for an example section 'a.2.c.+::d'
    - case      : the word boundary.                              = '.'
    - end       : the rhs end str.                                = '::'
    - types     : allowed types.                                  = str|int
    - marks     : StrangMarkAbstract_e of words with a meta meaning.  = '+'
    - required  : a strang errors if a required section isnt found

    - idx       : the index of the section

    TODO Maybe 'type_re' and 'mark_re'

    """
    __slots__ = ("case", "end", "idx", "marks", "name", "required", "types")
    idx       : int
    name      : Final[str]
    case      : Final[Maybe[str]]
    end       : Final[Maybe[str]]
    types     : Final[type|UnionType]
    marks     : Final[Maybe[type[StrangMarkAbstract_e]]]
    required  : Final[bool]

    def __init__(self, name:str, case:Maybe[str], end:Maybe[str], types:type|UnionType, marks:Maybe[type[StrangMarkAbstract_e]], required:bool=True, *, idx:int=-1) -> None:  # noqa: FBT001, FBT002, PLR0913
        assert(case is None or bool(case))
        assert(end is None or bool(end))
        self.idx       = idx
        self.name      = name.lower()
        self.case      = case
        self.end       = end
        self.types     = types
        self.marks     = marks
        self.required  = required

    def __contains__(self, other:type|StrangMarkAbstract_e) -> bool:
        match other:
            case type() as x:
                return issubclass(x, self.types)
            case UnionType() as xs:
                # Check its contained using its removal of duplicates
                return (xs | self.types) == self.types
            case StrangMarkAbstract_e() as x if self.marks:
                return x in self.marks
            case _:
                return False

class Sections_d:
    """
    An object to hold information about word separation and sections,
    a strang type is structured into these

    Each Section is a Sec_d
    TODO add format conversion specs
    """
    __slots__ = ("named", "order", "types")
    named  : Final[dict[str, int]]
    order  : Final[tuple[Sec_d, ...]]
    types  : type|UnionType

    def __init__(self, *sections:tuple|Sec_d) -> None:
        order : list[Sec_d] = []
        for i, sec in enumerate(sections):
            match sec:
                case Sec_d() as obj:
                    obj.idx = i
                    order.append(obj)
                case xs:
                    obj = Sec_d(*xs, idx=i)
                    order.append(obj)
        else:
            assert(all(x.end is not None for x in order[:-1])), SEC_END_MSG
            self.order = tuple(order)
            self.named = {x.name:i for i,x in  enumerate(self.order)}

    def __contains__(self, val:str) -> bool:
        return val in self.named

    def __getitem__(self, val:int|str) -> Sec_d:
        match val:
            case int() as i:
                return self.order[i]
            case str() as k if k in self.named:
                return self.order[self.named[k]]
            case x:
                raise KeyError(x)

    def __iter__(self) -> Iterator[Sec_d]:
        return iter(self.order)

    def __len__(self) -> int:
        return len(self.order)

class Strang_d:
    """ Extra Data of a Strang.
    Sections are accessed by their index, so use cls._sections.named[name] to get the index

    - sections  : tuple[slice, ...]      - Section boundaries
    - sec_words : tuple[tuple[int, ...]] - lookup of (sec, word) -> WordIndex
    - words     : tuple[slice, ...]      - Word Slices
    - meta      : tuple[Maybe, ...]      - Flat word level meta data

    """
    __slots__ = ("args", "args_start", "flat_idx", "meta", "sec_words", "sections", "uuid", "words")
    args_start  : Maybe[int]
    args        : Maybe[tuple]
    sections    : tuple[slice, ...]
    sec_words   : tuple[tuple[int, ...], ...]
    flat_idx    : tuple[tuple[int, int], ...]
    words       : tuple[slice, ...]
    meta        : tuple[Maybe, ...]
    uuid        : Maybe[UUID]

    def __init__(self, uuid:Maybe[UUID]=None) -> None:
        self.args_start  = None
        self.args        = None
        self.sections    = ()
        self.sec_words   = ()
        self.words       = ()
        self.meta        = ()
        self.uuid        = uuid


##--| Default Section Specs
HEAD_SEC             : Final[Sec_d]       = Sec_d("head", CASE_DEFAULT, END_DEFAULT, BODY_TYPES, DefaultHeadMarks_e, True)  # noqa: FBT003
BODY_SEC             : Final[Sec_d]       = Sec_d("body", CASE_DEFAULT, None, HEAD_TYPES, DefaultBodyMarks_e, True)  # noqa: FBT003

CODEREF_HEAD_SEC     : Final[Sec_d]       = Sec_d("head",   CASE_DEFAULT, END_DEFAULT, HEAD_TYPES, CodeRefHeadMarks_e, False)  # noqa: FBT003
CODEREF_MODULE_SEC   : Final[Sec_d]       = Sec_d("module", CASE_DEFAULT, ":", HEAD_TYPES, DefaultBodyMarks_e, True)  # noqa: FBT003
CODEREF_VAL_SEC      : Final[Sec_d]       = Sec_d("value",  CASE_DEFAULT, None, HEAD_TYPES, CodeRefHeadMarks_e, True)  # noqa: FBT003

STRANG_DEFAULT_SECS  : Final[Sections_d]  = Sections_d(HEAD_SEC, BODY_SEC)
STRANG_ALT_SECS      : Final[Sections_d]  = Sections_d(
    Sec_d("head", CASE_DEFAULT, END_DEFAULT, HEAD_TYPES, DefaultHeadMarks_e, True),  # noqa: FBT003
    Sec_d("body", CASE_DEFAULT, None, BODY_TYPES, DefaultBodyMarks_e, True),  # noqa: FBT003
)
CODEREF_DEFAULT_SECS : Final[Sections_d] = Sections_d(CODEREF_HEAD_SEC, CODEREF_MODULE_SEC, CODEREF_VAL_SEC)
##--| Protocols

@runtime_checkable
class Importable_p(Protocol):
    """  """
    pass

class StrangUUIDs_p(Protocol):

    def to_uniq(self, *args:str) -> Self: ...

    def de_uniq(self) -> Self: ...

class StrangMod_p(Protocol):

    def pop(self, *, top:bool=False) -> Strang_p: ...

    def push(self, *vals:PushVal) -> Strang_p: ...

class StrangFormatter_p(Protocol):
    """ A string.Formatter with some Strang-specific methods """

    def format(self, format_string:str, /, *args:Any, **kwargs:Any) -> str: ... # noqa: ANN401

    def get_value(self, key:str, args:Any, kwargs:Any) -> str:  ...  # noqa: ANN401

    def convert_field(self, value:Any, conversion:Any) -> str: ...  # noqa: ANN401

    def expanded_str(self, data:Strang_p, *, stop:Maybe[int]=None) -> str: ...

@runtime_checkable
class Strang_p(StrangUUIDs_p, StrangMod_p, String_p, Protocol):
    """  """
    _processor  : ClassVar[PreProcessor_p]
    _formatter  : ClassVar[string.Formatter]
    _sections   : ClassVar[Sections_d]
    data        : Strang_d

    @classmethod
    def sections(cls) -> Sections_d: ...

    @classmethod
    def section(cls, arg:int|str) -> Sec_d: ...

    @override
    def __getitem__(self, i:ItemIndex) -> str: ... # type: ignore[override]

    @property
    def base(self) -> Self: ...

    @property
    def shape(self) -> tuple[int, ...]: ...

    @override
    def index(self, *sub:FindSlice, start:Maybe[int]=None, end:Maybe[int]=None) -> int: ... # type: ignore[override]

    @override
    def rindex(self, *sub:FindSlice, start:Maybe[int]=None, end:Maybe[int]=None) -> int: ... # type: ignore[override]

    def words(self, idx:SectionIndex, *, case:bool=False) -> list: ...

    def get(self, *args:SectionIndex|WordIndex) -> Any: ...  # noqa: ANN401

    def args(self) -> Maybe[tuple]: ...
    def uuid(self) -> Maybe[UUID]: ...



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN202, B011
# imports:
from __future__ import annotations

import itertools as itz
import logging as logmod
import warnings
import pathlib as pl
import typing
from typing import Any
import pytest
from jgdv.cli import ParseError
from ..param_spec import ParamSpec
from ... import _interface as API  # noqa: N812
from .. import core
##--| vars
logging = logmod.root

good_names = ("test", "blah", "bloo")
bad_names  = ("-test", "blah=bloo")

##--|

class TestToggleParam:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_fail_with_wrong_type(self):
        data     = {"name" : "-test", "type":str}
        with pytest.raises(TypeError):
            core.ToggleParam(**data)

    def test_consume_toggle(self):
        data     = {"name" : "-test"}
        in_data  = [ "-test" ]
        obj      = core.ToggleParam(**data)
        match obj.consume(in_data):
            case {"test": True}, 1:
                assert(True)
            case x:
                assert(False), x

    def test_consume_inverse_toggle(self):
        data     = {"name" : "-test"}
        in_data  = ["-no-test"]
        obj      = core.ToggleParam(**data)
        assert(obj.default_value is False)
        match obj.consume(in_data):
            case {"test": False}, 1:
                assert(True)
            case x:
                assert(False), x

    def test_consume_short_toggle(self):
        data     = {"name" : "-test"}
        in_data  = ["-t"]
        obj      = core.ToggleParam(**data)
        match obj.consume(in_data):
            case {"test": True}, 1:
                assert(True)
            case x:
                assert(False), x

class TestPositionalParam:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_protocol(self):
        assert(issubclass(core.PositionalParam, ParamSpec))
        assert(isinstance(core.PositionalParam, API.ParamStruct_p))
        assert(issubclass(core.PositionalParam, API.PositionalParam_p))

    def test_consume_positional(self):
        data     = {"name":"<1>test", "type":str}
        in_data  = ["aweg", "blah"]
        obj      = core.PositionalParam(**data)
        match obj.consume(in_data):
            case {"test": "aweg"}, 1:
                assert(True)
            case x:
                assert(False), x

    def test_consume_positional_list(self):
        data = {
            "name"     : "test",
            "type"     : list,
            "default"  : [],
            "count"    : 2,
        }
        in_data = ["bloo", "blah", "aweg"]
        obj = core.PositionalParam(**data)
        match obj.consume(in_data):
            case {"test": ["bloo", "blah"]}, 2:
                assert(True)
            case x:
                assert(False), x


    def test_consume_count_unrestricted(self):
        data = {
            "name"     : "test",
            "type"     : list,
            "default"  : [],
            "count"    : -1,
        }
        in_data = ["bloo", "blah", "aweg", "aweg", "qqqq"]
        obj = core.PositionalParam(**data)
        match obj.consume(in_data):
            case {"test": list() as vals}, int() as count:
                assert(count == len(in_data))
                for x,y in zip(vals, in_data, strict=True):
                    assert(x == y)

            case x:
                assert(False), x


    def test_consume_unrestricted_one_still_is_list(self):
        data = {
            "name"     : "test",
            "type"     : list,
            "default"  : [],
            "count"    : -1,
        }
        in_data = ["bloo"]
        obj = core.PositionalParam(**data)
        match obj.consume(in_data):
            case {"test": list() as vals}, int() as count:
                assert(count == len(in_data))
                for x,y in zip(vals, in_data, strict=True):
                    assert(x == y)

            case x:
                assert(False), x

class TestKeyParam:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_consume_key_value_str(self):
        data     = {"name" : "-test", "type":str}
        in_data  = ["-test", "blah"]
        obj      = core.KeyParam(**data)
        assert(obj.type_ is str)
        match obj.consume(in_data):
            case {"test":"blah"}, 2:
                assert(True)
            case x:
                assert(False), x

    def test_consume_key_value_int(self):
        data     = {"name" : "-test", "type": int}
        in_data  = ["-test", "20"]
        obj      = core.KeyParam(**data)
        assert(obj.type_ is int)
        match obj.consume(in_data):
            case {"test":20}, 2:
                assert(True)
            case x:
                assert(False), x

    def test_consume_key_value_fail(self):
        data     = {"name" : "-test", "type":str}
        in_data  = ["-nottest", "blah"]
        obj      = core.KeyParam(**data)
        match obj.consume(in_data):
            case None:
                assert(True)
            case _:
                assert(False)

class TestAssignParam:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_fail_with_wrong_type(self):
        data     = {"name" : "-test=", "type":bool}
        with pytest.raises(TypeError):
            core.AssignParam(**data)

    def test_fail_with_no_prefix(self):
        data     = {"name" : "test="}
        with pytest.raises(ValueError):
            core.AssignParam(**data)

    def test_fail_with_no_separator(self):
        data     = {"name" : "-test"}
        with pytest.raises(ValueError):
            core.AssignParam(**data)

    def test_build_with_custom_prefix(self):
        data     = {"name" : "+test="}
        val = core.AssignParam(**data)
        assert(val.prefix == "+")

    def test_consume_assignment(self):
        data     = {"name" : "--test="}
        in_args  = ["--test=blah", "other"]
        obj      = core.AssignParam(**data)
        match obj.consume(in_args):
            case {"test":"blah"}, 1:
                assert(True)
            case x:
                assert(False), x

    def test_consume_int(self):
        data     = {"name" : "--test=", "type":int}
        in_args  = ["--test=2", "other"]
        obj      = core.AssignParam(**data)
        match obj.consume(in_args):
            case {"test": 2}, 1:
                assert(True)
            case x:
                assert(False), x

    def test_consume_assignment_wrong_prefix(self):
        data     = {"name" : "--test="}
        in_args  = ["-t=blah"]
        obj      = core.AssignParam(**data)
        match obj.consume(in_args):
            case None:
                assert(True)
            case x:
                assert(False), x

class TestLiteralParam:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_literal(self):
        obj = core.LiteralParam(name="blah")
        match obj.consume(["blah"]):
            case {"blah":True}, 1:
                assert(True)
            case None:
                assert(False)

    def test_literal_fail(self):
        obj = core.LiteralParam(name="blah")
        match obj.consume(["notblah"]):
            case None:
                assert(True)
            case _:
                assert(False)




#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202, B011, PLR2004

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

from jgdv.structs.strang import Strang
from jgdv.mixins.annotate._interface import Default_K
from ..dkey import DKey
from ..keys import SingleDKey, NonDKey

# ##-- types
# isort: off
import abc
import collections.abc
import types
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
   from jgdv import Maybe
   from typing import Final
   from typing import ClassVar, Any, LiteralString
   from typing import Never, Self, Literal
   from typing import TypeGuard
   from collections.abc import Iterable, Iterator, Callable, Generator
   from collections.abc import Sequence, Mapping, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class TestBaseDKey:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_registry_default(self):
        assert(DKey._registry[Default_K] is SingleDKey)

    def test_class_getitem(self):
        val = DKey['blah']
        assert(val is not None)
        assert(isinstance(val, types.GenericAlias))

    def test_basic_implicit(self):
        match DKey("blah", implicit=True):
            case NonDKey():
                assert(False)
            case DKey() as obj:
                assert(not hasattr(obj, "__dict__"))
                assert(isinstance(obj, str))
                assert(isinstance(obj, Strang))
                assert(hasattr(obj, "__hash__"))
            case x:
                assert(False), x

    def test_basic_explicit(self):
        match DKey("{blah}"):
            case NonDKey():
                assert(False)
            case DKey() as obj:
                assert(not hasattr(obj, "__dict__"))
                assert(isinstance(obj, str))
                assert(isinstance(obj, Strang))
                assert(hasattr(obj, "__hash__"))
            case x:
                assert(False), x

    def test_expansion_limit_format_params(self):
        match DKey("blah:e2", implicit=True):
            case DKey() as obj:
                assert(bool(obj.data.raw[0].format))
                assert(obj.data.raw[0].format == "e2")
                assert(obj.data.max_expansions == 2)
            case x:
                assert(False), x


    def test_multikey_expansion_limit_format_params(self):
        obj = DKey("{blah:e2} {bloo:e3}")
        match obj.keys():
            case [x,y]:
                assert(x.data.max_expansions == 2)
                assert(y.data.max_expansions == 3)
            case x:
                assert(False), x

    def test_hashable_implicit(self):
        obj = DKey("blah", implicit=True)
        assert(hash(obj))

    def test_hashable_explicit(self):
        obj = DKey("{blah}")
        assert(hash(obj))

    def test_eq(self):
        obj1 = DKey("blah", implicit=True)
        obj2 = DKey("blah", implicit=True)
        assert(obj1 == obj2)

    def test_eq_str(self):
        obj1 = DKey("blah", implicit=True)
        obj2 = "blah"
        assert(obj1 == obj2)

    def test_eq_not_implemented(self):
        obj1 = DKey("blah", implicit=True)
        obj2 = 21
        assert(not (obj1 == obj2))

class TestDKey_Format:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_format_indirect(self):
        obj = DKey("{blah}")
        assert(f"{obj:i}" == "blah_")

    def test_format_direct(self):
        obj = DKey("{blah}")
        assert(f"{obj:d}" == "blah")

    def test_format_wrapped(self):
        obj = DKey("{blah}")
        assert(f"{obj:w}" == "{blah}")

    def test_format_wrapped_indirect(self):
        obj = DKey("{blah}")
        assert(f"{obj:wi}" == "{blah_}")



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN202, B011, PLR2004
# Imports
from __future__ import annotations

##-- stdlib imports
import logging as logmod
import pathlib as pl
from typing import (Any, ClassVar, Generic, TypeAlias,
                    TypeVar, cast)
from collections.abc import Callable, Iterable, Iterator, Mapping, MutableMapping, Sequence
from re import Match
from types import GenericAlias
import warnings

##-- end stdlib imports

import pytest
from .. import Subclasser, SubAlias_m

# Logging:
logging = logmod.root

# Global Vars:

class TestSubAlias:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_simple(self):

        class Basic(SubAlias_m, fresh_registry=True):
            pass

        assert(hasattr(Basic, "_registry"))
        assert(hasattr(Basic, "_clear_registry"))
        assert(Basic.cls_annotation() == ())

    def test_matching_direct(self):

        class Basic(SubAlias_m, fresh_registry=True):
            pass

        class IntBasic(Basic[int]):
            pass

        inst = IntBasic()
        match inst:
            case IntBasic():
                assert(True)
            case x:
                assert(False), x

    def test_matching_superclass(self):

        class Basic(SubAlias_m, fresh_registry=True):
            pass

        class IntBasic(Basic[int]):
            pass

        inst = IntBasic()
        match inst:
            case Basic():
                assert(True)
            case x:
                assert(False), x

class TestSubAlias_Access:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_class_getitem_generic_aliases(self):

        class Basic(SubAlias_m, fresh_registry=True):
            pass

        match Basic[int]:
            case GenericAlias() as x:
                assert(x.__args__ == (int,))
            case x:
                assert(False), x

        match Basic[float]:
            case GenericAlias() as x:
                assert(x.__args__ == (float,))
            case x:
                assert(False), x

        match Basic["test"]:
            case GenericAlias() as x:
                assert(x.__args__ == ("test",))
            case x:
                assert(False), x

    def test_repeated_access(self):

        class Basic(SubAlias_m, fresh_registry=True):
            pass

        gen1 = Basic[int]
        gen2 = Basic[int]
        assert(gen1 is not gen2)
        assert(gen1 == gen2)

class TestSubAlias_Registry:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_fresh_registry(self):

        class Basic(SubAlias_m):
            pass

        class Other(SubAlias_m, fresh_registry=True):
            pass

        assert(Basic._registry is not Other._registry)

class TestSubAlias_Subclassing:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_annotated(self):

        class Basic(SubAlias_m, fresh_registry=True):
            pass

        class SimpleBasic(Basic):
            pass

        assert(issubclass(SimpleBasic, Basic))
        assert(SimpleBasic.cls_annotation() is Basic.cls_annotation())
        assert(SimpleBasic.cls_annotation() == ())

    def test_unannotated_repeat(self):

        class Basic(SubAlias_m, fresh_registry=True):
            pass

        class SimpleBasic(Basic):
            pass

        class SubSimple(SimpleBasic):
            pass

        assert(issubclass(SimpleBasic, Basic))
        assert(SimpleBasic.cls_annotation() is Basic.cls_annotation())
        assert(SimpleBasic.cls_annotation() == ())

    def test_unannotated_multi(self):

        class Basic(SubAlias_m, fresh_registry=True):
            pass

        class SimpleBasic(Basic):
            pass

        class OtherBasic(Basic):
            pass

        assert(issubclass(SimpleBasic, Basic))
        assert(SimpleBasic.cls_annotation() is Basic.cls_annotation())
        assert(SimpleBasic.cls_annotation() == ())

    def test_annotated_by_access(self):

        class Basic(SubAlias_m, fresh_registry=True):
            pass

        class IntBasic(Basic[int]):
            pass

        assert(IntBasic is Basic[int])
        assert(issubclass(IntBasic, Basic))
        assert(IntBasic.cls_annotation() is Basic[int].cls_annotation())
        assert(IntBasic.cls_annotation() == (int,))


    def test_annotated_by_kwarg(self):

        class Basic(SubAlias_m, fresh_registry=True):
            pass

        class IntBasic(Basic, annotation=int):
            pass

        assert(IntBasic is Basic[int])
        assert(issubclass(IntBasic, Basic))
        assert(IntBasic.cls_annotation() is Basic[int].cls_annotation())
        assert(IntBasic.cls_annotation() == (int,))


    def test_multi_annotation_by_kwarg(self):

        class Basic(SubAlias_m, fresh_registry=True):
            pass

        class IntBasic(Basic, annotation=[int, "blah"]):
            pass

        assert(IntBasic is Basic[int, "blah"])
        assert(issubclass(IntBasic, Basic))
        assert(IntBasic.cls_annotation() is Basic[int, "blah"].cls_annotation())
        assert(IntBasic.cls_annotation() == (int,"blah"))

    def test_retrieved_subclass_conflict_error(self):

        class Basic(SubAlias_m, fresh_registry=True, strict=True):
            pass

        class IntBasic(Basic[int]):
            pass

        assert(IntBasic.cls_annotation() == (int,))
        with pytest.raises(TypeError):

            class RetrievedIntBasic(Basic[int]):
                pass

    def test_explicit_subclass_conflict_error(self):

        class Basic(SubAlias_m, fresh_registry=True, strict=True):
            pass

        class IntBasic(Basic[int]):
            pass

        with pytest.raises(TypeError):

            class ExplicitIntBasic(IntBasic):
                pass

class TestSubAlias_Accumulation:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_accumulation_a(self):

        class Basic(SubAlias_m, fresh_registry=True, accumulate=True):
            pass

        class IntBasic(Basic[int]):
            pass

        class AnotherIntBasic(Basic[int]['another']):
            pass

        assert(issubclass(IntBasic, Basic))
        assert(issubclass(AnotherIntBasic, Basic))
        assert(issubclass(AnotherIntBasic, IntBasic))
        assert(AnotherIntBasic.cls_annotation() == (int, "another"))

    def test_accumulation_b(self):

        class Basic(SubAlias_m, fresh_registry=True, accumulate=True):
            pass

        class IntBasic(Basic[int]):
            pass

        class AnotherIntBasic(IntBasic['blah']):
            pass

        assert(issubclass(IntBasic, Basic))
        assert(issubclass(AnotherIntBasic, Basic))
        assert(issubclass(AnotherIntBasic, IntBasic))
        assert(AnotherIntBasic.cls_annotation() == (int, "blah"))



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from . import errors

# ##-- types
# isort: off
import abc
import collections.abc
import typing
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    import pathlib as pl
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from ._interface import Strang_p
    from jgdv._abstract.pre_processable import PreProcessor_p, PreProcessResult, InstanceData, PostInstanceData
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
StrMeta      : Final[type]  = type(str)
HasDictFail  : Final[str]   = "The resulting strang has a __dict__. Set the subclass to have __slots__=()"
# Body:

class StrangMeta(StrMeta):
    """ A Metaclass for Strang
    It runs the pre-processsing and post-processing on the constructed str
    to turn it into a strang
    """

    def __call__[T:Strang_p](cls:type[T], text:str|pl.Path, *args:Any, **kwargs:Any) -> Strang_p:  # noqa: ANN401, N805
        """ Overrides normal str creation to allow passing args to init """
        ctor       : type[T]
        obj        : T
        processor  : PreProcessor_p[T]  = cls._processor
        stage      : str                = "Pre-Process"

        try:
            text, inst_data, post_data, new_ctor = processor.pre_process(cls,
                                                                         text,
                                                                         *args,
                                                                         strict=kwargs.pop("strict", False),
                                                                         **kwargs,
                                                                         )
            ctor   = new_ctor or cls
            assert(isinstance(ctor, type|typing.GenericAlias)), ctor # type: ignore[attr-defined]
            stage  = "__new__"
            obj    = ctor.__new__(ctor, text)
            stage  = "__init__"
            obj.__init__(*args, **collections.ChainMap(inst_data, kwargs)) # type: ignore[misc]
            stage  = "Process"
            obj    = processor.process(obj, data=post_data) or obj
            stage  = "Post-Process"
            obj    = processor.post_process(obj, data=post_data) or obj
        except TypeError as err:
            raise errors.StrangError(errors.StrangCtorFailure.format(cls=cls.__name__, stage=stage),
                                     err, text, cls, processor) from None
        except ValueError as err:
            raise errors.StrangError(errors.StrangCtorFailure.format(cls=cls.__name__, stage=stage),
                                     err, text, cls, processor) from None
        else:
            assert(isinstance(obj, str))
            if hasattr(obj, "__dict__"):
                raise ValueError(HasDictFail, type(obj))
            return obj




"""

jgdv.debugging : Utilities for debugging

Provides:
- SignalHandler      : for installing handlers for interrupts
- TimeBlock_ctx      : CtxManager for simple timing
- MultiTimeBlock_ctx : for more complicated timing
- TraceBuilder       : for slicing the traceback provided in exceptions
- LogDel             : a class decorator for logging when __del__ is called

"""
from .signal_handler import SignalHandler, NullHandler
from .trace_builder import TraceBuilder
from .log_del import LogDel



#!/usr/bin/env python3
"""
A Prelude of imports to use


"""
# Import:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from uuid import UUID, uuid1
# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

# isort: on
# ##-- end types

from .types import *



#!/usr/bin/env python3
"""

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import abc
import sys
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
from random import randint
import re
import time
import types
import zipfile
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Final,
    Generator,
    Generic,
    Iterable,
    Iterator,
    Mapping,
    Match,
    MutableMapping,
    Protocol,
    Sequence,
    Tuple,
    TypeAlias,
    TypeGuard,
    TypeVar,
    cast,
    final,
    overload,
    runtime_checkable,
)
from uuid import UUID, uuid1
from weakref import ref

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv import Maybe

# ##-- end 1st party imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging


zip_name_default         : Final[str]                   = "default"
zip_overwrite_default    : Final[bool]                  = False
zip_compression_default  : Final[str]                   = "ZIP_DEFLATED"
zip_level_default        : Final[int]                   = 4

zip_choices              : Final[list[tuple[str, str]]] = [
    ("none", "No compression"),
    ("zip", "Default Zip Compression"),
    ("bzip2", "bzip2 Compression"),
    ("lzma", "lzma compression")
]

class Zipper_m:
    """
    Add methods for manipulating zip files.
    Can set a self.zip_root path, where added files with be relative to
    """
    zip_name            : str                   = zip_name_default
    zip_overwrite       : bool                  = zip_overwrite_default
    zip_root            : Maybe[pl.Path]        = None
    _zip_compression    : str                   = zip_compression_default
    _zip_compress_level : int                   = zip_level_default

    def _zip_get_compression_settings(self) -> tuple[int, int]:
        match self.args:
            case { "compression": "none", "level": x }:
                return zipfile.ZIP_STORED, x
            case { "compression": "zip", "level": x }:
                return zipfile.ZIP_DEFLATED, x
            case { "compression": "bzip2", "level": x }:
                return zipfile.ZIP_BZIP2, x
            case { "compression" : "lzma", "level": x}:
                return zipfile.ZIP_LZMA, x
            case _:
                return self._zip_compression, self._zip_compress_level

    def zip_set_root(self, fpath:pl.Path):
        """ set the filesystem that acts as the root for paths to be added to the zip file """
        self.zip_root = fpath

    def zip_create(self, fpath:pl.Path):
        """ Create a new zipfile. will overwrite an existing zip if 'zip_overwrite' is set """
        assert(fpath.suffix== ".zip")
        if self.zip_overwrite and fpath.exists():
            fpath.unlink()
        elif fpath.exists():
            return

        logging.info("Creating Zip File: %s", fpath)
        now = datetime.datetime.strftime(datetime.datetime.now(), "%Y:%m:%d::%H:%M:%S")
        record_str = f"Zip File created at {now} for doot task: {self.basename}"
        compress_type, compress_level = self._zip_get_compression_settings()

        with zipfile.ZipFile(fpath, mode='w', compression=compress_type, compresslevel=compress_level, allowZip64=True ) as targ:
            targ.writestr(".taskrecord", record_str)

    def zip_add_paths(self, fpath:pl.Path, *args:pl.Path):
        """
        Add specific files to the zip.
          Will Create the zip if it doesn't exist
        """
        logging.info("Adding to Zipfile: %s : %s", fpath, args)
        assert(fpath.suffix == ".zip")
        self.zip_create(fpath)

        root = self.zip_root or pl.Path()
        paths = [pl.Path(x) for x in args]
        compress_type, compress_level = self._zip_get_compression_settings()
        with zipfile.ZipFile(fpath, mode='a', compression=compress_type, compresslevel=compress_level, allowZip64=True ) as targ:
            for file_to_add in paths:
                try:
                    relpath = file_to_add.relative_to(root)
                    attempts = 0
                    write_as = relpath
                    while str(write_as) in targ.namelist():
                        if attempts > 10:
                            logging.warning(f"Couldn't settle on a de-duplicated name for: {file_to_add}")
                            break
                        logging.debug(f"Attempted Name Duplication: {relpath}", file=sys.stderr)
                        write_as = relpath.with_stem(f"{relpath.stem}_{hex(randint(1,100))}")
                        attempts += 1

                    targ.write(str(file_to_add), write_as)

                except ValueError:
                    relpath = root / pl.Path(file_to_add).name
                except FileNotFoundError as err:
                    logging.warning(f"Adding File to Zip {fpath} failed: {err}", file=sys.stderr)

    def zip_globs(self, fpath:pl.Path, *globs:str, ignore_dots=False):
        """
        Add files chosen by globs to the zip, relative to the cwd
        """
        logging.debug("Zip Globbing: %s : %s", fpath, globs)
        assert(fpath.suffix == ".zip")
        self.zip_create(fpath)

        root                          = self.zip_root or pl.Path()
        compress_type, compress_level = self._zip_get_compression_settings()
        with zipfile.ZipFile(fpath, mode='a', compression=compress_type, compresslevel=compress_level, allowZip64=True) as targ:
            for globstr in globs:
                result = list(root.glob(globstr))
                logging.info(f"Globbed: {root}/{globstr} : {len(result)}")
                for globf in result:
                    try:
                        if globf.stem[0] == "." and ignore_dots:
                            continue
                        relpath = pl.Path(globf).relative_to(root)
                        match str(relpath) in targ.namelist():
                            case True:
                                logging.warning("Duplication Attempt: %s -> %s", globf, relpath)
                            case False:
                                targ.write(str(globf), relpath)
                    except FileNotFoundError as err:
                        logging.warning(f"Adding File to Zip {fpath} failed: {err}", file=sys.stderr)

    def zip_add_str(self, fpath:pl.Path, fname:str, text:str):
        """ add a string of text to a zip file as a new file """
        assert(fpath.suffix == ".zip")
        self.zip_create(fpath)

        compress_type, compress_level = self._zip_get_compression_settings()
        with zipfile.ZipFile(fpath, mode='a', compression=compress_type, compresslevel=compress_level, allowZip64=True) as targ:
            match fname in targ.namelist():
                case True:
                    logging.warning("Duplication Attempt: %s -> %s", fpath, fname)
                case False:
                    targ.writestr(fname, text)


    def zip_get_contents(self, fpath:pl.Path) -> list[str]:
        with zipfile.Zipfile(fpath):
            return zipfile.namelist()


    def zip_unzip_to(self, fpath:pl.Path, *zips:pl.Path, fn=None):
        """
        extract everything or everything that returns true from fn, from all zips given
        into subdirs of fpath
        """
        fn = fn or (lambda x: True)

        for zipf in zips:
            logging.debug("Extracting: %s (%s) to %s", zipf, fn, fpath)
            (fpath / zipf.stem).mkdir(parents=True, exist_ok=True)
            with zipfile.ZipFile(zipf) as targ:
                subset = [x for x in targ.namelist() if fn(x)]
                targ.extractall(fpath / zipf.stem, members=subset)

    def zip_unzip_concat(self, fpath:pl.Path, *zips:pl.Path, member=None, header=b"\n\n#------\n\n", footer=b"\n\n#------\n\n"):
        """ Unzip a member file in a multiple zip files,
          append their text contents into a single file """
        assert(member is not None)
        with open(fpath, "ab") as out:
            for zipf in zips:
                try:
                    logging.debug("Concating: %s (%s) to %s", zipf, member, fpath)
                    with zipfile.ZipFile(zipf) as targ:
                        data = targ.read(member)
                        if header:
                            out.write(header)
                        out.write(data)
                        if footer:
                            out.write(footer)

                except Exception as err:
                    logging.warning("Issue reading: %s : %s", zipf, err)

    def zip_test(self, *zips:pl.Path):
        """ Test the validity of zip files """
        for zipf in zips:
            with zipfile.ZipFile(zipf) as targ:
                result = targ.testzip()
                if result is not None:
                    logging.warning("Issue with %s : %s", zipf, targ)


    def zip_contains(self, zip:pl.Path, *names:str|pl.Path) -> bool:
        """ test that a zip file contains multiple filenames"""
        with zipfile.ZipFile(zip, "r") as zipf:
            contents = zipf.namelist()

        missing = [x for x in names if x not in contents]
        if bool(missing):
            logging.info("Zip file %s is missing : %s", zip, missing)

        return bool(missing)



"""

"""
# mypy: disable-error-code="attr-defined"
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from collections import ChainMap
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from statemachine import State, StateMachine
from statemachine.exceptions import TransitionNotAllowed
from statemachine.states import States

# ##-- end 3rd party imports

# ##-- 1st party imports
from jgdv import Proto
from jgdv.structs.chainguard import ChainGuard

# ##-- end 1st party imports

from . import errors
from .param_spec.param_spec import ParamSpec, ParamProcessor
from .param_spec import HelpParam, SeparatorParam
from .parse_machine_base import ParseMachineBase
from . import _interface as API # noqa: N812
from ._interface import ParseResult_d, EXTRA_KEY, EMPTY_CMD

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from dataclasses import InitVar, dataclass, field

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|
from ._interface import ParamStruct_p, ParamStruct_i, ArgParser_p, ParamSource_p
# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

HELP            : Final[ParamSpec]     = HelpParam()
SEPERATOR       : Final[ParamSpec]     = SeparatorParam()

class ParseMachine(ParseMachineBase):
    """ Implemented Parse State Machine

    __call__ with:
    args       : list[str]       -- the cli args to parse (ie: from sys.argv)
    head_specs : list[ParamSpec] -- specs of the top level program
    cmds       : list[ParamSource_p] -- commands that can provide their parameters
    subcmds    : dict[str, list[ParamSource_p]] -- a mapping from commands -> subcommands that can provide parameters

    A cli call will be of the form:
    {proghead} {prog args} {cmd} {cmd args}* [{subcmd} {subcmdargs} [-- {subcmd} {subcmdargs}]* ]? (--help)?

    eg:
    doot -v list -by-group a b c --help
    doot run basic::task -quick --value=2 --help

    Will raise a jgdv.cli.errors.ParseError on failure
    """

    def __init__(self, **kwargs) -> None:  # noqa: ANN003
        kwargs.setdefault("parser", CLIParser())
        super().__init__(**kwargs)

    def __call__(self, args:list[str], *, head_specs:list[ParamSpec], cmds:list[ParamSource_p], subcmds:list[tuple[str, ParamSource_p]]) -> Maybe[dict]:
        assert(self.current_state == self.Start) # type: ignore[has-type]
        try:
            self.setup(args, head_specs, cmds, subcmds)
            if self.current_state not in [self.ReadyToReport, self.Failed, self.End]: # type: ignore[has-type]
                self.parse()
            self.finish()
        except TransitionNotAllowed as err:
            msg = "Transition failure"
            raise errors.ParseError(msg, err) from err
        else:
            result = self.model.report()
            # Reset the state
            self.current_state = self.Start
            return result

@Proto(ArgParser_p)
class CLIParser:
    """

    # {prog} {args} {cmd} {cmd_args}
    # {prog} {args} [{task} {tasks_args}] - implicit do cmd

    """
    _initial_args      : list[str]
    _remaining_args    : list[str]
    _head_specs        : list[ParamSpec]
    _cmd_specs         : dict[str, list[ParamStruct_i]]
    _subcmd_specs      : dict[str, tuple[str|tuple[str, ...], list[ParamStruct_i]]]
    head_result        : Maybe[ParseResult_d]
    cmd_result         : Maybe[ParseResult_d]
    subcmd_results     : list[ParseResult_d]
    extra_results      : ParseResult_d
    _force_help        : bool

    def __init__(self) -> None:
        self._remaining_args    = []
        self._initial_args      = []
        self._remaining_args    = []
        self._head_specs        = []
        self._cmd_specs         = {}
        self._subcmd_specs      = {}
        self.head_result        = None
        self.cmd_result         = None
        self.subcmd_results     = []
        self.extra_results      = ParseResult_d(EXTRA_KEY)
        self._force_help        = False

    def _parse_fail_cond(self) -> bool:
        return False

    def _has_no_more_args_cond(self) -> bool:
        return not bool(self._remaining_args)

    @ParseMachineBase.finish._transitions.before
    def all_args_consumed_val(self) -> None:
        if bool(self._remaining_args):
            msg = "Not All Args Were Consumed"
            raise errors.ArgParseError(msg, self._remaining_args)

    @ParseMachineBase.Prepare.enter
    def _setup(self, args:list[str], head_specs:list, cmds:list[ParamSource_p], subcmds:list[tuple[str, ParamSource_p]]) -> None:
        """
          Parses the list of arguments against available registered parameter head_specs, cmds, and tasks.
        """
        logging.debug("Setting up Parsing : %s", args)
        head_specs            = head_specs or []
        cmds                  = cmds or []
        subcmds               = subcmds or []
        self._initial_args    = args[:]
        self._remaining_args  = args[:]
        self._head_specs      = head_specs

        self.head_result       = None
        self.cmd_result        = None
        self.subcmd_results    = []
        self.extra_results     = ParseResult_d(EXTRA_KEY)
        self._force_help       = False

        if not isinstance(cmds, list):
            msg = "cmds needs to be a list"
            raise TypeError(msg, cmds)

        ##--| Prep cmd lookup
        ##
        for x in cmds:
            match x:
                case (str() as alias, ParamSource_p() as source):
                    self._cmd_specs[alias] = source.param_specs()
                case ParamSource_p():
                    self._cmd_specs[x.name] = x.param_specs()
                case x:
                    raise TypeError(x)

        ##--| Prep subcmd lookup
        ## maps {subtask} -> (cmds, params)
        match subcmds:
            case [*xs]:
                self._subcmd_specs = {y.name:(x, y.param_specs()) for x,y in xs}
            case _:
                logging.info("No Subcmd Specs provided for parsing")
                self._subcmd_specs = {}


    @ParseMachineBase.Cleanup.enter
    def _cleanup(self) -> None:
        logging.debug("Cleaning up")
        self._initial_args      = []
        self._remaining_args    = []
        self._cmd_specs         = {}
        self._subcmd_specs      = {}

    @ParseMachineBase.CheckForHelp.enter
    def help_flagged(self) -> None:
        logging.debug("Checking for Help Flag")
        match HELP.consume(self._remaining_args[-1:]):
            case None:
                self._force_help = False
            case _:
                self._force_help = True
                self._remaining_args.pop()

    @ParseMachineBase.Head.enter
    def _parse_head(self) -> None:
        """ consume arguments for doot actual """
        logging.debug("Head Parsing: %s", self._remaining_args)
        if not bool(self._head_specs):
            self.head_result = ParseResult_d(name=self._remaining_args.pop(0))
            return
        head_specs       = sorted(self._head_specs, key=ParamSpec.key_func)
        defaults : dict  = ParamSpec.build_defaults(head_specs)
        self.head_result = ParseResult_d("_head_", defaults)
        self._parse_params_unordered(self.head_result, head_specs)

    @ParseMachineBase.Cmd.enter
    def _parse_cmd(self) -> None:
        """ consume arguments for the command being run """
        cmd_specs  : list[ParamStruct_i]
        defaults   : dict
        cmd_name   : str
        logging.debug("Cmd Parsing: %s", self._remaining_args)
        if not bool(self._cmd_specs):
            self.cmd_result = ParseResult_d(EMPTY_CMD, {})
            return

        # Determine cmd
        cmd_name = self._remaining_args[0]
        if cmd_name not in self._cmd_specs:
            self.cmd_result = ParseResult_d(EMPTY_CMD, {})
            return

        logging.info("Cmd matches: %s", cmd_name)
        self._remaining_args.pop(0)
        # get its specs
        cmd_specs = sorted(self._cmd_specs[cmd_name], key=ParamSpec.key_func)
        defaults  = ParamSpec.build_defaults(cmd_specs)
        self.cmd_result                 = ParseResult_d(cmd_name, defaults)
        self._parse_params_unordered(self.cmd_result, cmd_specs) # type: ignore

    @ParseMachineBase.SubCmd.enter
    def _parse_subcmd(self) -> None:
        """ consume arguments for tasks """
        active_cmd  : str
        last        : Maybe[str]
        sub_specs   : list[ParamStruct_i]
        defaults    : dict
        sub_result  : ParseResult_d

        if not bool(self._subcmd_specs):
            return

        logging.debug("SubCmd Parsing: %s", self._remaining_args)
        assert(self.cmd_result is not None)
        active_cmd  = self.cmd_result.name
        last        = None
        # Determine subcmd
        while (bool(self._remaining_args)
               and last != (sub_name:=self._remaining_args[0])
               ):
            if self._parse_separator():
                continue
            else:
                self._remaining_args.pop(0)

            logging.debug("Sub Cmd: %s", sub_name)
            last = sub_name
            match self._subcmd_specs.get(sub_name, None):
                case str() as cmd_constraint, list() as params if active_cmd in [cmd_constraint, EMPTY_CMD]:
                    sub_specs   = sorted(params, key=ParamSpec.key_func)
                    defaults    =  ParamSpec.build_defaults(sub_specs)
                    sub_result  = ParseResult_d(sub_name, defaults)
                    self._parse_params_unordered(sub_result, sub_specs) # type: ignore
                    self.subcmd_results.append(sub_result)
                case [*constraints], list() as params if active_cmd in constraints or active_cmd == EMPTY_CMD:
                    sub_specs   = sorted(params, key=ParamSpec.key_func)
                    defaults    =  ParamSpec.build_defaults(sub_specs)
                    sub_result  = ParseResult_d(sub_name, defaults)
                    self._parse_params_unordered(sub_result, sub_specs) # type: ignore
                    self.subcmd_results.append(sub_result)

                case _, _:
                    pass
                case _:
                    msg = "Unrecognised SubCmd"
                    raise errors.SubCmdParseError(msg, sub_name)
        else:
            logging.debug("SubCmds Parsed")

    @ParseMachineBase.Extra.enter
    def _parse_extra(self) -> None:
        logging.debug("Extra Parsing: %s", self._remaining_args)
        self._remaining_args = []

    def _parse_params_unordered(self, res:ParseResult_d, params:list[ParamSpec]) -> None:
        """ For a given list of parameters, consume as many as possible.
        First by consuming non-positional (ie: key and key-val params)
        Followed by Positional (params with relative ordering using <int>).

        Expects the params to be sorted already.
        """
        x : Any
        logging.debug("Parsing Params Unordered: %s", params)
        processor       : ParamProcessor       = ParamSpec._processor
        non_positional  : list[ParamStruct_i]  = []
        positional      : list[ParamStruct_i]  = []

        for x in params:
            if isinstance(x, API.PositionalParam_p):
                positional.append(cast("ParamStruct_i", x))
            else:
                non_positional.append(x)

        def consume_it(x:ParamStruct_i) -> None:
            """ Local function to parse and remove args from the remainder """
            # TODO refactor this as a partial
            match x.consume(self._remaining_args):
                case None:
                    msg = "Failed to consume"
                    raise errors.ParseError(msg, x.name)
                case dict() as data, int() as count:
                    logging.debug("Consumed: %s/%s", count, len(self._remaining_args))
                    self._remaining_args = self._remaining_args[count:]
                    res.args.update(data)
                    res.non_default.add(x.name)
                case y:
                    raise TypeError(type(y))

        # Parse non-positional params
        logging.debug("- Starting non-positional arg parsing: %s", len(non_positional))
        while bool(non_positional) and bool(self._remaining_args):
            match [x for x in non_positional if processor.matches_head(x, self._remaining_args[0], silent=False)]:
                case []:
                    non_positional = []
                case [x]:
                    consume_it(x)
                    non_positional.remove(x)
                case [*xs]:
                    msg = "Too many potential non-positional params"
                    raise errors.ParseError(msg, xs)
        else:
            logging.debug("- Finished consuming non-positional")

        # Parse positional params
        logging.debug("- Starting positional arg parsing: %s", len(positional))
        while bool(positional) and bool(self._remaining_args):
            curr = positional[0]
            if processor.matches_head(curr, self._remaining_args[0], silent=False):
                consume_it(curr)

            positional.remove(curr)
        else:
            logging.debug("- Finished Consuming Positional")
            logging.debug("Remaining args: %s", len(self._remaining_args))

    def _parse_separator(self) -> bool:
        match SEPERATOR.consume(self._remaining_args):
            case None:
                return False
            case _:
                logging.debug("----")
                self._remaining_args.pop(0)
                return True

    def report(self) -> Maybe[dict]:
        """ Take the parsed results and return a nested dict """

        match self._force_help:
            case False:
                cmd_result = self.cmd_result
            case True if self.cmd_result is None:
                assert(self.head_result is not None)
                self.head_result.args['help'] = True
                cmd_result = ParseResult_d(EMPTY_CMD)
            case True if self.cmd_result.name == EMPTY_CMD:
                cmd_result = ParseResult_d("help", {"target":None, "args": self.cmd_result.args}, self.cmd_result.non_default)
            case True:
                cmd_result = ParseResult_d("help", {"target":self.cmd_result.name, "args": self.cmd_result.args}, self.cmd_result.non_default)

        result = {
            "head"  : self.head_result.to_dict() if self.head_result else {},
            "cmd"   : cmd_result.to_dict() if cmd_result else {},
            "sub"   : {y['name']:y['args'] for x in self.subcmd_results if (y:=x.to_dict()) is not None},
            "extra" : self.extra_results.to_dict(),
        }
        return result



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN001, ANN002, ANN003
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import builtins
import datetime
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv._abstract.pre_processable import PreProcessor_p
from jgdv.mixins.annotate import SubAlias_m
from jgdv.structs.strang import _interface as StrangAPI  # noqa: N812
from jgdv.structs.strang.processor import StrangBasicProcessor

# ##-- end 1st party imports

from . import _interface as API  # noqa: N812
from ._interface import DKeyMark_e, Key_p
from ._util._interface import ExpInst_d
from ._util.parser import DKeyParser

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never, ClassVar
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError
from collections.abc import Mapping

if TYPE_CHECKING:
   from jgdv import Maybe, Ident, Ctor
   import enum
   from typing import Final
   from typing import ClassVar, Any, LiteralString
   from typing import Never, Self, Literal
   from typing import TypeGuard
   from collections.abc import Sized
   from collections.abc import Iterable, Iterator, Callable, Generator
   from collections.abc import Sequence, MutableMapping, Hashable
   from string import Formatter

   from jgdv._abstract.pre_processable import PreProcessResult
   from ._interface import KeyMark

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--| Body:

class DKeyProcessor[T:API.Key_p](PreProcessor_p):
    """
      The Metaclass for keys, which ensures that subclasses of DKeyBase
      are API.Key_p's, despite there not being an actual subclass relation between them.

    This allows DKeyBase to actually bottom out at str
    """

    parser               : ClassVar[DKeyParser]    = DKeyParser()
    _expected_init_keys  : ClassVar[list[str]]     = API.DEFAULT_DKEY_KWARGS[:]

    expected_kwargs      : Final[list[str]]        = API.DEFAULT_DKEY_KWARGS
    convert_mapping      : dict[str, KeyMark]
    ##--|

    def __init__(self) -> None:
        self.convert_mapping = {}

    @override
    def pre_process(self, cls:type[T], input:Any, *args:Any, strict:bool=False, **kwargs:Any) -> PreProcessResult[T]: # type: ignore[override]  # noqa: PLR0912, PLR0915
        """ Pre-process the Key text,

        Extracts subkeys, and refines the type of key to build
        """
        text         : str
        ctor         : Ctor[T]
        mark         : API.KeyMark
        spec_mark    : API.KeyMark
        format_mark  : Maybe[API.KeyMark]
        ##--|
        inst_data    : dict            = {}
        post_data    : dict            = {}
        force        : Maybe[Ctor[T]]  = kwargs.pop('force', None)
        implicit     : bool            = kwargs.pop("implicit", False)  # is key wrapped? ie: {key}
        insist       : bool            = kwargs.pop("insist", False)    # must produce key, not nullkey

        # TODO handle generic aliases by using the arg as instance expansion type
        match force, kwargs.pop('mark', None):
            case type(), _:
                spec_mark = cls.MarkOf(force)
            case None, None:
                spec_mark = cls.MarkOf(cls)
            case None, x:
                spec_mark = x
            case _:
                spec_mark = cls.MarkOf(cls)

        ##--|
        if spec_mark is None:
            spec_mark = API.DKeyMark_e.default()
        # TODO use class hook if it exists

        ##--| Pre-clean text
        match input:
            case str():
                text = input.strip()
            case pl.Path():
                text = str(input).strip()
            case _:
                text = str(input).strip()

        # Early exit if the text is empty:
        if not bool(text):
            inst_data['mark'] = None
            ctor = self.select_ctor(cls, mark=False, force=None, insist=False)
            return str(input), inst_data, post_data, ctor

        ##--| Get pre-parsed keys
        match kwargs.pop(API.RAWKEY_ID, None) or self.extract_raw_keys(text, implicit=implicit):
            case [x]:
                inst_data[API.RAWKEY_ID] = [x]
            case [*xs]:
                inst_data[API.RAWKEY_ID] = xs
            case []:
                inst_data[API.RAWKEY_ID] = []
            case x:
                msg = "No raw keys were able to be extracted"
                raise TypeError(msg, type(x))

        ##--| discriminate on raw keys
        match self.inspect_raw(inst_data[API.RAWKEY_ID], kwargs):
            case x, y:
                text = x or text
                format_mark = y
            case x:
                msg = "Inspecting raw keys failed"
                raise ValueError(msg, x)
        ##--|
        match spec_mark, format_mark:
            case type() as x, _:
                mark = x
            case x, None:
                mark = x
            case x, y if x == y:
                mark = x
            case x, y if y == DKeyMark_e.null() and x != DKeyMark_e.multi():
                assert(y is not None)
                mark = y
            case x, y if x is DKeyMark_e.default():
                assert(y is not None)
                mark = y
            case str() as x, _ if x not in DKeyMark_e:
                mark = x
            case _, y:
                assert(y is not None)
                mark = y


        assert(bool(text))
        assert(bool(inst_data))
        ctor = self.select_ctor(cls, insist=insist, mark=mark, force=force)
        self.validate_init_kwargs(ctor, kwargs)
        assert(issubclass(ctor, SubAlias_m))
        inst_data['mark'] = ctor.cls_annotation()
        if DKeyMark_e.multi() in [format_mark, spec_mark] and not issubclass(ctor, API.MultiKey_p):
            msg = "a multi key was specified by a mark, but the ctor isnt a multi key"
            raise ValueError(msg, spec_mark, format_mark)
        ##--| return
        return text, inst_data, post_data, ctor

    @override
    def process(self, obj:T, *, data:Maybe[dict]=None) -> Maybe[T]:
        """ The key constructed, build slices """
        # TODO use class hook if it exists
        full        : str
        wrapped     : bool
        start       : int
        stop        : int
        key_slices  : list[slice]
        raw_keys    : list[API.RawKey_d]
        if not bool(obj.data.raw):  # Nothing to do
            return None

        key_slices  = []
        raw_keys    = []
        wrapped     = API.OBRACE in obj[:]
        if wrapped and 1 < len(obj.data.raw):
            raw_keys += obj.data.raw

        for key in raw_keys:
            if not key.key:
                continue

            full = key.joined()
            if wrapped:
                full = f"{{{full}}}"

            start = obj.index(full)
            stop  = start + len(full)
            key_slices.append(slice(start, stop))

        else:
            # TODO Add a word slice for each sub key
            obj.data.sec_words  = (tuple(range(len(obj.data.raw))),) # tuple[tuple[slice, ...]]
            obj.data.words      = tuple(key_slices)
            obj.data.flat_idx   = tuple((i,j) for i,x in enumerate(obj.data.sec_words) for j in range(len(x)))
            obj.data.sections   = (slice(0, len(obj)),) # a single, whole str slice

            return None

    @override
    def post_process(self, obj:T, data:Maybe[dict]=None) -> Maybe[T]:
        """ Build subkeys if necessary

        """
        # for each subkey, build it...
        x : Any
        key_meta : list[Maybe[str|API.Key_p]] = []
        raw : list[API.RawKey_d] = []
        if isinstance(obj, API.MultiKey_p):
            raw = obj.data.raw

        for x in raw:
            key_meta.append(x.joined())
        else:
            obj.data.meta = tuple(key_meta)

        match getattr(obj, "_post_process_h", None):
            case hook if  callable(hook):
                hook(data)
            case None:
                pass
            case x:
                raise TypeError(type(x))

        return None
    ##--| Utils

    def inspect_raw(self, raw_keys:Iterable[API.RawKey_d], kdata:dict) -> tuple[Maybe[str], Maybe[API.KeyMark]]:  # noqa: ARG002
        """ Take extracted keys of the text,
        and determine features of them.
        can return modified text, and a mark

        """
        assert(all(isinstance(x, API.RawKey_d) for x in raw_keys))
        format_mark  : Maybe[API.KeyMark]  = None
        text         : Maybe[str]          = None
        match raw_keys:
            case [x] if not bool(x.key) and bool(x.prefix): # No keys found, use NullDKey
                format_mark  = False
            case [x] if not bool(x.prefix):  # One key, no non-key text. trim it.
                if x.convert and x.convert in self.convert_mapping:
                    format_mark = self.convert_mapping[x.convert]
                if x.is_indirect():
                    format_mark = Mapping
                text = x.direct()
            case [_, *_]: # Multiple keys found, coerce to multi
                format_mark = list
            case []: # No Keys
                pass
            case x:
                msg = "Unrecognised raw keys type"
                raise TypeError(msg, type(x))
        ##--|
        return text, format_mark

    def select_ctor(self, cls:Ctor[T], *, mark:KeyMark, force:Maybe[Ctor[T]], insist:bool) -> Ctor[T]:
        """ Select the appropriate key ctor,
        which can be forced if necessary,
        otherwise uses the mark and multi params

        """
        # Choose the sub-ctor
        assert(issubclass(cls, SubAlias_m))
        if force is not None:
            assert(isinstance(force, type)), force
            return force

        try:
            match cls._retrieve_subtype(mark):
                case types.GenericAlias() as x:
                    return x
                case type() as ctor if insist and ctor.MarkOf(ctor) is cls.Marks.null():
                    raise TypeError(API.InsistentKeyFailure)
                case type() as x:
                    return cast("type[T]", x)
        except KeyError:
            return cls
        else:
            return cls

    def extract_raw_keys(self, data:str, *, implicit=False) -> tuple[API.RawKey_d, ...]:
        """ Calls the Python format string parser to extract
        keys and their formatting/conversion specs,
        then wraps them in jgdv.structs.dkey._util.parser.API.RawKey_d's for convenience

        if 'implicit' then will parse the entire string as {str}
        """
        return tuple(self.parser.parse(data, implicit=implicit))

    def consume_format_params(self, spec:str) -> tuple[str, bool, bool]:
        """
          return (remaining, wrap, direct)
        """
        wrap     = 'w' in spec
        indirect = 'i' in spec
        direct   = 'd' in spec
        remaining = API.FMT_PATTERN.sub("", spec)
        assert(not (direct and indirect))
        return remaining, wrap, (direct or (not indirect))

    def validate_init_kwargs(self, ctor:type[Key_p], kwargs:dict) -> None:
        """ returns any keys not expected by a dkey or dkey subclass """
        assert(ctor is not None)
        result = set(kwargs.keys() - self.expected_kwargs - ctor._extra_kwargs)
        if bool(result):
            raise ValueError(API.UnexpectedKwargs, result)


    def register_convert_param(self, cls:type[API.Key_p], convert:Maybe[str]) -> None:
        match convert:
            case str() as x if x in self.convert_mapping:
                msg = "Convert Mapping Already Registered"
                raise KeyError(msg, x, self.convert_mapping[x])
            case str() as x:
                self.convert_mapping[x] = cls.MarkOf(cls) # type: ignore[arg-type]
            case _:
                pass



#!/usr/bin/env python3
"""

"""

##-- builtin imports
from __future__ import annotations

import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1
import tracemalloc
import linecache
import fnmatch

##-- end builtin imports

from . import _interface as API # noqa: N812
from jgdv import Proto, Mixin
from jgdv.mixins.human_numbers import HumanNumbers_m

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Traceback
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Snapshot   = tracemalloc.Snapshot
    type Statistic  = tracemalloc.Statistic
    type Difference = tracemalloc.StatisticDiff
    type Logger     = logmod.Logger
    type Filter     = tracemalloc.Filter
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--| Vals
##--| Body
@Mixin(HumanNumbers_m)
class MallocTool:
    """ see https://docs.python.org/3/library/tracemalloc.html

    example::

        with MallocTool(2) as dm:
            dm.whitelist(__file__)
            dm.blacklist("*tracemalloc.py", all_frames=False)
            val = 2
            dm.snapshot("simple")
            vals = [random.random() for x in range(1000)]
            dm.current()
            dm.snapshot("list")
            vals = None
            dm.current()
            dm.snapshot("cleared")

        dm.compare("simple", "list")
        dm.compare("list", "cleared")
        dm.compare("list", "simple")
        dm.inspect("list")

    """
    num_frames      : int
    started         : bool
    snapshots       : list[Snapshot]
    named_snapshots : dict[str, Snapshot]
    filters         : list[Filter]
    _logger         : Maybe[Logger]
    _log_level      : int

    def __init__(self, *, num_frames:int=5, logger:Maybe[Logger]=None, level:int=API.DEFAULT_LOG_LEVEL) -> None:
        self.num_frames      = num_frames
        self.started         = False
        self.snapshots       = []
        self.named_snapshots = {}
        self.filters         = []
        self._logger         = logger
        self._log_level      = level

    def __enter__(self) -> Self:
        tracemalloc.start(self.num_frames)
        self.started = True
        self.snapshot(name="init")
        return self

    def __exit__(self, etype:Maybe[type], err:Maybe[Exception], tb:Maybe[Traceback]) -> bool: #type:ignore[exit-return]
        self.snapshot(name="final")
        self.started = False
        self._log(f"Recorded {len(self.snapshots)} snapshots")
        return False  # type: ignore[exit-return]

    def whitelist(self, file_pat:str, lineno:Maybe[int]=None, *, all_frames:bool=True) -> None:
        self.filters.append(
            tracemalloc.Filter(inclusive=True,
                               filename_pattern=file_pat,
                               lineno=lineno,
                               all_frames=all_frames),
        )

    def blacklist(self, file_pat:str, *, lineno:Maybe[int]=None, all_frames:bool=True) -> None:
        self.filters.append(
            tracemalloc.Filter(inclusive=False,
                               filename_pattern=file_pat,
                               lineno=lineno,
                               all_frames=all_frames),
            )

    def file_matches(self, name:str|pl.Path, pat:str) -> bool:
        match name:
            case str():
                return fnmatch.fnmatch(name, pat)
            case pl.Path():
                return fnmatch.fnmatch(str(name), pat)
            case x:
                raise TypeError(type(x))

    def snapshot(self, *, name:Maybe[str]=None) -> None:
        if not self.started:
            msg = "DebugMalloc needs to have been entered"
            raise RuntimeError(msg)

        snap = tracemalloc.take_snapshot()
        self.snapshots.append(snap)
        if name and name not in self.named_snapshots:
            self.named_snapshots[name] = snap

        tracemalloc.clear_traces()

    def get_snapshot(self, val:int|str) -> Snapshot:
        match val:
            case int() if 0 <= val < len(self.snapshots):
                snap = self.snapshots[val]
            case int() if val < 0:
                snap = self.snapshots[val]
            case str() if val in self.named_snapshots:
                snap = self.named_snapshots[val]
            case _:
                raise TypeError(val)

        return snap.filter_traces(self.filters)

    def current(self, val:Maybe=None) -> None:
        traced = tracemalloc.get_traced_memory()
        curr_mem = self.humanize(traced[0])  # type: ignore[attr-defined]
        peak_mem = self.humanize(traced[1])  # type: ignore[attr-defined]
        self._log("Current Memory: %s, Peak: %s", curr_mem, peak_mem)
        if val:
            self._log("Value allocated at: %s", tracemalloc.get_object_traceback(val))

    def inspect(self, val:Any, *, type:str=API.DEFAULT_REPORT) -> None:  # noqa: A002, ANN401
        self._log(f"\n-- Inspecting {val}")
        snap = self.get_snapshot(val)
        for stat in snap.statistics(type):
            self._print_stat(stat)

    def compare(self, val1:int|str, val2:int|str, *, type:str=API.DEFAULT_REPORT) -> None:  # noqa: A002
        self._log(f"\n-- Comparing {val1} to {val2}")
        snap1 = self.get_snapshot(val1)
        snap2 = self.get_snapshot(val2)

        differences = snap2.compare_to(snap1, type)
        diff_count  = len(differences)
        for i, stat in enumerate(differences):
            self._log(f"- {val1} -> {val2} diff {i}/{diff_count}:")
            self._print_diff(stat)

    def _print_diff(self, stat:Difference) -> None:
        """ Print a Trace memory comparison """
        assert(isinstance(stat, tracemalloc.StatisticDiff))
        size_diff  = self.humanize(stat.size_diff, force_sign=True)  # type: ignore[attr-defined]
        count_diff = stat.count_diff
        self._log("%s, %s blocks", size_diff, count_diff, prefix=API.CHANGE_PREFIX)
        self._print_stat(stat)

    def _print_stat(self, stat:Statistic|Difference) -> None:
        """ Print a Traced memory snapshot """
        assert(isinstance(stat, tracemalloc.Statistic|tracemalloc.StatisticDiff))
        tb = stat.traceback
        self._log(f"Frame/Count/Size: {len(stat.traceback):}, {stat.count:}, {self.humanize(stat.size)}")  # type: ignore[attr-defined]
        for x in range(len(tb) -1):
            frame = tb[x]
            path = pl.Path(frame.filename)
            self._log(f"  {path.name:<15}", "line:", f"{frame.lineno:<7}: {linecache.getline(frame.filename, frame.lineno).strip()}")

        frame = tb[-1]
        path = pl.Path(frame.filename)
        self._log(f"* {path.name:<15}", "line:", f"{frame.lineno:<7}: {linecache.getline(frame.filename, frame.lineno).strip()}")

    def _log(self, msg:str, *args:Any, prefix:str=API.DEFAULT_PREFIX) -> None:  # noqa: ANN401
        if self._logger is None:
            return None
        if not bool(msg):
            self._logger.log(self._log_level, "")
            return

        full_fmt = f"{prefix} {msg}"
        self._logger.log(self._log_level, full_fmt, *args)




#!/usr/bin/env python3
"""

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import typing
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

from jgdv.debugging import TraceBuilder
from ._core import MonotonicDec, IdempotentDec, Decorator, DataDec
from jgdv.mixins.annotate import Subclasser

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, TypeAliasType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from types import resolve_bases, FunctionType, MethodType

if TYPE_CHECKING:
    from ._interface import Decorated
    from ._interface import Decorable, DForm_e
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--| Global Vars:
MIXIN_KWD : Final[str] = "_jgdv_mixins"
ABSMETHS  : Final[str] = "__abstractmethods__"
IS_ABS    : Final[str] = "__isabstractmethod__"
NAME_MOD  : Final[str] = "M"
##--| Funcs

##--| Body

class Mixin(MonotonicDec):
    """ Decorator to App/Prepend Mixins into the decorated class.

    Converts::

        class ClsName(mixins, Supers, Protocols, metaclass=MCls, **kwargs):...

    into::

        @Protocols(*ps)
        @Mixin(*ms, None)
        class ClsName(Supers): ...


    ('None' is used to separate pre and post mixins)
"""

    __builder  : ClassVar[Subclasser]  = Subclasser()
    needs_args                         = True

    def __init__(self, *mixins:Maybe[type], allow_inheritance:bool=False, silent:bool=False) -> None:
        super().__init__()
        self._silent = silent
        self._name_mod = NAME_MOD
        try:
            index = mixins.index(None)
        except ValueError:
            index = 0
        finally:
            self._pre_mixins  = [x for x in mixins[:index] if x is not None and x is not object]
            self._post_mixins = [x for x in mixins[index:] if x is not None and x is not object]
            if not allow_inheritance:
                self._validate_mixins()


    def _validate_mixins(self) -> None:
        for x in itz.chain(self._pre_mixins, self._post_mixins):
            if len(x.mro()) > 3:  # noqa: PLR2004
                msg = "Can't Mixin a class that inherits anything"
                raise TypeError(msg, x)

    def _build_annotations_h(self, target:Decorable, current:list) -> Maybe[list]:  # noqa: ARG002
        """ Given a list of the current annotation list,
        return its replacement
        """
        new_mixins = current[:]
        new_mixins  = [x for x in self._pre_mixins if x not in current]
        new_mixins += [x for x in self._post_mixins  if x not in current]
        return new_mixins

    def _validate_target_h(self, target:Decorable, form:DForm_e, args:Maybe[list]=None) -> None:  # noqa: ARG002
        match target:
            case type() if hasattr(target, "model_fields"):
                msg = "Pydantic classes shouldn't be extended"
                raise TypeError(msg, target)
            case type():
                pass
            case _:
                msg = "Unexpected type passed for mixin annotation"
                raise TypeError(msg, target)

    def _wrap_class_h(self, cls:Decorable) -> Decorated:
        assert(isinstance(cls, type))
        match cls.mro():
            case [x, *xs]:
                new_mro = [*self._pre_mixins, x, *self._post_mixins, *xs]
            case _:
                pass
        match self._silent:
            case False:
                new_name  = self.__builder.decorate_name(cls, self._name_mod)
            case True:
                new_name = cls.__name__

        mixed     = self.__builder.make_subclass(new_name, cls, mro=new_mro)
        self.annotate_decorable(mixed)
        return mixed


class DelayMixin(DataDec):
    """ TODO A Decorator for annotating a class with mixins

    Delays the construction of the True class until later,
    using @MixinNow
    """
    pass

class MixinNow(MonotonicDec):
    """ TODO The trigger for delayed mixins.

    After using @DelayMixin,
    trigger the True class using this.

    eg::

        @MixinNow
        @DelayMixin(m3, None, m4)
        @DelayMixin(m1, m2)
        class Blah:...

    """
    pass
